/*

Repair -- an implementation of Larsson and Moffat's compression and
decompression algorithms.
Copyright (C) 2010-current_year Gonzalo Navarro

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author's contact: Gonzalo Navarro, Dept. of Computer Science, University of
Chile. Blanco Encalada 2120, Santiago, Chile. gnavarro@dcc.uchile.cl

*/

// linear probing hash table for pairs

	// table only grows to keep factor, does not shrink
	// this is ok for this application where #pairs rarely decreases

	// value -1 denotes empty cells, -2 is a deletion mark

#include <stdlib.h>
#include <stdio.h>
#include "Hash.h"

//extern float factor; // load factor

typedef unsigned long long relong;
#define LPRIME ((relong)767865341467865341)
#define PRIME 2013686449


Hash::Hash(int _maxpos, float _factor)// creates new empty hash table
  {
    int i;
	// upgrade maxpos to the next value of the form (1<<smth)-1
    while (_maxpos & (_maxpos-1)) _maxpos &= _maxpos-1;
    _maxpos = (_maxpos-1)<<1 | 1;  // avoids overflow if maxpos = 1<<31
    maxpos = _maxpos;
    used = 0;
    table = new int[1+maxpos];
    for (i=0;i<=maxpos;i++) table[i] = -1;
    factor = _factor;
  }

int Hash::searchHash (Tpair p, Trecord *recs) // returns id

  { relong u = ((relong)p.left)<<(8*sizeof(int)) | (relong)p.right;
    int k = ((LPRIME*u) >> (8*sizeof(int))) & maxpos;
    while (table[k] != -1)
      {	if ((table[k] >= 0) &&
	    (recs[table[k]].pair.left == p.left) &&
	    (recs[table[k]].pair.right == p.right)) break;
	k = (k+1) & maxpos;
      }
    return table[k];
  }

void Hash::deleteHash (int id, Trecord *rec) // deletes H->Rec[id].pair from hash

  { table[rec[id].kpos] = -2;
    used--;
  }


 int Hash::finsertHash (Tpair p)
			// inserts w/o resizing, assumes there is space
			// does not update used field
			// note can reuse marked deletions

  { relong u = ((relong)p.left)<<(8*sizeof(int)) | (relong)p.right;
    int k = ((LPRIME*u) >> (8*sizeof(int))) & maxpos;
    while (table[k] >= 0) k = (k+1) & maxpos;
    return k;
  }

void Hash::insertHash (int id, Trecord *rec) // inserts H->Rec[id].pair in hash
				  // assumes key is not present
				  // sets ptr from Rec to hash as well

  { int k;
    if (used > maxpos * factor) // resize
	{ Hash *newH = new Hash((maxpos<<1)|1, factor);
	  int i;
	  int *tab = table;
	  for (i=0;i<=maxpos;i++)
	      if (tab[i] >= 0) // also removes marked deletions
		 { k = newH->finsertHash (rec[tab[i]].pair);
		   newH->table[k] = tab[i];
		   rec[tab[i]].kpos = k;
		 }
	  newH->used = used;
	  delete [] table;
	  used = newH->used;
          table = newH->table;
          maxpos = newH->maxpos;
          //Rec = newH->Rec;
	}
    used++;
    k = finsertHash (rec[id].pair);
    table[k] = id;
    rec[id].kpos = k;
  }

Hash::~Hash() {
    delete [] table;
    table = NULL;
    maxpos = 0;
    used = 0;
  }

void Hash::hashRepos (int id, Trecord *rec )

  { table[rec[id].kpos] = id;
  }



