/*

Repair -- an implementation of Larsson and Moffat's compression and
decompression algorithms.
Copyright (C) 2010-current_year Gonzalo Navarro

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author's contact: Gonzalo Navarro, Dept. of Computer Science, University of
Chile. Blanco Encalada 2120, Santiago, Chile. gnavarro@dcc.uchile.cl

*/


	// binary heap with sqrt(u) heaps for least occurring ones

#include <stdlib.h>
#include <stdio.h>
#include "HeapBal.h"



HeapBal::HeapBal(int u, float factor, int minsize, Records_Bal *_Rec)
				// creates new empty heap
				// minsize, factor: space/time tradeoffs
  {
    int i;
    sqrtu = 2;
    while (sqrtu * sqrtu < u) sqrtu++;
    infreq = new array_RPBal*[sqrtu];
    for (i=1;i<sqrtu;i++){
        infreq[i] = new array_RPBal(_Rec, factor, minsize);
    }
    freq = new Thnode [sqrtu];
    freef = 0;
    for (i=0;i<sqrtu-1;i++) freq[i].next = i+1;
    freq[sqrtu-1].next = -1;
    ff = new Thfreq [sqrtu];
    freeff = 0;
    for (i=0;i<sqrtu-1;i++) ff[i].larger = i+1;
    ff[sqrtu-1].larger = -1;
    smallest = largest = -1;
    max = sqrtu;
    //Rec = _Rec;
  }


static void move (array_RPBal *A, int i, int j, Trecord *rec)

  { int id = A->pairs[j];
    A->pairs[i] = id;
    rec[id].hpos = i;
  }

void HeapBal::prnH ()

  { Thfreq *f;
    static int X = 0;
    int prevf = 1<<30;
    int fp = largest;
    if (fp == -1) return;
    X++;
    printf ("HeapBal %i = \n",X);
    while (fp != -1)
       { f = &ff[fp];
         printf ("freq=%i, elems=%i\n",f->freq,f->elems);
	 if (prevf <= f->freq)
	    { fp++; }
	 prevf = f->freq;
	 fp = f->smaller;
       }
  }

void HeapBal::incFreq (int id, Trecord *rec) // inc freq of pair Rec[id]

  { int _freq = rec[id].freq++;
    int hpos = rec[id].hpos;
    Thnode *p;
    Thfreq *f,*lf;
    int fp,lfp;
if (PRNH) prnH();
    if (_freq >= sqrtu) // high freq part, hpos is a ptr within freq
       { p = &freq[hpos];
	 fp = p->fnode;
	 f = &ff[fp];
	 _freq++;
		// shortcut for common case: f has only p and !exists f+1
	 if ((p->prev == -1) && (p->next == -1) &&
	     ((f->larger == -1) || (ff[f->larger].freq != _freq)))
	    { f->freq = _freq; }
	 else  // the long way, hopefully not so common
	    {   	// remove p from f list
	      if (p->prev == -1) f->elems = p->next;
	      else freq[p->prev].next = p->next;
	      if (p->next != -1) freq[p->next].prev = p->prev;
		     // add p to larger list (lf)
	      p->prev = -1;
	      if ((f->larger != -1) && (ff[f->larger].freq == _freq))
						// next freq exists
	         { lfp = f->larger;
		   lf = &ff[lfp];
	           p->next = lf->elems;
		   freq[lf->elems].prev = hpos;
	         }
	      else // create one
	         { lfp = freeff;
	           lf = &ff[lfp];
		   freeff = lf->larger;
	           lf->freq = _freq;
	           lf->smaller = fp; lf->larger = f->larger;
		   if (f->larger != -1) ff[f->larger].smaller = lfp;
		   else largest = lfp;
	           f->larger = lfp;
	           p->next = -1;
	         }
	      lf->elems = hpos;
	      p->fnode = lfp;
		     // see if f disappears
	      if (f->elems == -1)
	         { if (f->smaller == -1) smallest = f->larger;
	           else ff[f->smaller].larger = f->larger;
	           if (f->larger == -1) largest = f->smaller;
	           else ff[f->larger].smaller = f->smaller;
	           f->larger = freeff;
		   freeff = fp;
	         }
	    }
       }
    else // del from low freq part
       { move (infreq[_freq], hpos, infreq[_freq]->fst, rec);
	 infreq[_freq]->deleteArray();
	 _freq++;
         if (_freq < sqrtu) // ins in low freq part
            { rec[id].hpos = infreq[_freq]->insertArray (id);
            }
	 else // ins in freq part
	    { 		// allocate a free cell for it
	      hpos = freef;
	      freef = freq[freef].next;
	      rec[id].hpos = hpos;
	      p = &freq[hpos];
	      p->prev = -1;
	      p->id = id;
	      if ((smallest != -1) && (ff[smallest].freq == _freq))
							// freq exists
		 { fp = smallest;
		   f = &ff[fp];
		   p->next = f->elems;
		   freq[f->elems].prev = hpos;
		 }
	      else // create freq
	         { fp = freeff;
		   f = &ff[fp];
		   freeff = f->larger;
	           f->freq = _freq;
	           f->smaller = -1;
		   f->larger = smallest;
		   if (smallest != -1) ff[smallest].smaller = fp;
		   smallest = fp;
		   if (largest == -1) largest = fp;
	           p->next = -1;
	         }
	      f->elems = hpos;
	      p->fnode = fp;
	    }
       }
  }



void HeapBal::decFreq (int id, Records_Bal *Rec) // dec freq of pair Rec[id]

  { Trecord *rec = Rec->records;
    int _freq = rec[id].freq--;
    int hpos = rec[id].hpos;
    Thnode *p;
    Thfreq *f,*sf;
    int fp,sfp;
if (PRNH) prnH();
    if (_freq > sqrtu) // high freq part
       { p = &freq[hpos];
	 fp = p->fnode;
	 f = &ff[p->fnode];
	 _freq--;
		// shortcut for common case: f has only p and !exists f-1
	 if ((p->prev == -1) && (p->next == -1) &&
	     ((f->smaller == -1) || (ff[f->smaller].freq != _freq)))
	    { f->freq = _freq; }
	 else  // the long way, hopefully not so common
	    {   	// remove p from f list
	      if (p->prev == -1) f->elems = p->next;
	      else freq[p->prev].next = p->next;
	      if (p->next != -1) freq[p->next].prev = p->prev;
		     // add p to smaller list (sf)
	      p->prev = -1;
	      if ((f->smaller != -1) && (ff[f->smaller].freq == _freq))
							// next freq exists
	         { sfp = f->smaller;
	           sf = &ff[sfp];
	           p->next = sf->elems;
		   freq[sf->elems].prev = hpos;
	         }
	      else // create one
	         { sfp = freeff;
		   sf = &ff[sfp];
		   freeff = sf->larger;
	           sf->freq = _freq;
	           sf->larger = fp; sf->smaller = f->smaller;
		   if (f->smaller != -1) ff[f->smaller].larger = sfp;
		   else smallest = sfp;
	           f->smaller = sfp;
	           p->next = -1;
	         }
	      sf->elems = hpos;
	      p->fnode = sfp;
		     // see if f disappears
	      if (f->elems == -1)
	         { if (f->smaller == -1) smallest = f->larger;
	           else ff[f->smaller].larger = f->larger;
	           if (f->larger == -1) largest = f->smaller;
	           else ff[f->larger].smaller = f->smaller;
	           f->larger = freeff;
		   freeff = fp;
	         }
	    }
       }
    else // ins in low freq part
       { if (_freq < sqrtu) // del from low freq part
            { move (infreq[_freq], hpos, infreq[_freq]->fst,rec);
	      infreq[_freq]->deleteArray();
	    }
	 else // del from heap, must be minimal
	    {   	// remove p from f list
              p = &freq[hpos];
	      fp = p->fnode;
	      f = &ff[p->fnode];
	      if (p->prev == -1) f->elems = p->next;
	      else freq[p->prev].next = p->next;
	      if (p->next != -1) freq[p->next].prev = p->prev;
			// add its cell to free list
	      p->next = freef;
	      freef = hpos;
		     // see if f disappears
	      if (f->elems == -1)
	         { smallest = f->larger;
	           if (f->larger == -1) largest = -1;
	           else ff[f->larger].smaller = -1;
	           f->larger = freeff;
		   freeff = fp;
	         }
	    }
	 if (--_freq > 0) // freq 0 disappears from pairs
         {  rec[id].hpos = infreq[_freq]->insertArray (id);
         }else Rec->removeRecord(id);
       }
  }

void HeapBal::insertHeap (int id, Trecord *rec)  // with freq 1

  { rec[id].hpos = infreq[1]->insertArray(id);
    rec[id].freq = 1;
  }

int HeapBal::extractMax ()

  { //Trecord *rec = Rec->records;
    int ret;
    Thnode *p;
    Thfreq *f;
    int fp;
if (PRNH) prnH();
    if ((max == sqrtu) && (largest == -1)) max--;
    if (max < sqrtu)
       { while (max && (infreq[max]->size == 0)) max--;
	 if (!max) return -1; // empty heap
         ret = infreq[max]->pairs[infreq[max]->fst];
         infreq[max]->deleteArray();
       }
    else
       { fp = largest;
	 f = &ff[fp];
	 p = &freq[f->elems];
         ret = p->id;
		// remove element
	 f->elems = p->next;
	 if (p->next != -1) freq[p->next].prev = -1;
	 else { // remove f as well
	        largest = f->smaller;
	        if (f->smaller == -1) smallest = -1;
	        else ff[f->smaller].larger = -1;
	        f->larger = freeff;
		freeff = fp;
	      }
		// add elem to free list
	 p->next = freef;
	 freef = p-freq;
       }
    return ret;
  }



void HeapBal::purgeHeap (Records_Bal *Rec)
			// remove elems with freq 1 from heap and hash
			// their freq cannot grow after a repair turn
  {
    int id,fst,size,max;
    size = infreq[1]->size;
    fst = infreq[1]->fst;
    max = infreq[1]->maxsize;
    while (size--)
	{ id = infreq[1]->pairs[fst];
          fst = (fst+1) % max;
          Rec->removeRecord (id);
	}
    infreq[1]->~array_RPBal();
  }



void HeapBal::heapRepos (int id, Trecord *rec) // repositions pair

  { if (rec[id].freq < sqrtu)
         infreq[rec[id].freq]->pairs[rec[id].hpos] = id;
    else freq[rec[id].hpos].id = id;
  }

HeapBal::~HeapBal() // destroys H

  { int i;
    for (i=1;i<sqrtu;i++) infreq[i]->~array_RPBal();
    free (infreq); infreq = NULL;
    free (freq); freq = NULL;
    free (ff); ff = NULL;
    smallest = largest = -1;
    sqrtu = 0;
  }

