/*

Repair -- an implementation of Larsson and Moffat's compression and
decompression algorithms.
Copyright (C) 2010-current_year Gonzalo Navarro

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author's contact: Gonzalo Navarro, Dept. of Computer Science, University of
Chile. Blanco Encalada 2120, Santiago, Chile. gnavarro@dcc.uchile.cl

*/


	// binary heap with sqrt(u) heaps for least occurring ones
	// and list of frequencies, each with a list of elems, for the others
	// which are no more than sqrt(u) overall. this guarantees O(1)
	// operation times for the heap, as frequencies change by +-1, and
	// O(sqrt(u)) integers overhead over the bare arrays of ids (/factor).

#ifndef _HEAPBAL_H
#define	_HEAPBAL_H

#include <stdlib.h>
//#include "basics.h"
#include "array_RPBal.h"
#include "Records_Bal.h"

static int PRNH;


typedef struct _Thfreq
  { int freq;
    int elems; // a pointer within freq array
    int larger,smaller; // pointers within ff array
  } Thfreq;

typedef struct _Thnode
  { int id;
    int prev,next; // actually pointers within freq array
    int fnode; // ptr to its freq node (ptr to ff)
  } Thnode;

class array_RPBal;

class HeapBal{

public:

    HeapBal( int u, float factor, int minsize,Records_Bal *_Rec);
				// creates new empty heap
				// 0<factor<1: occupancy factor
				// sqrt(u)*max(minsize,n/factor) integers
    void incFreq ( int id, Trecord *rec) ; // inc freq of pair Rec[id]
    void decFreq (int id, Records_Bal *Rec); // dec freq of pair Rec[id]
    void insertHeap ( int id, Trecord *rec);  // with freq 1
    int extractMax ();
    void purgeHeap (Records_Bal * Rec); // remove elems with freq 1
    void heapRepos (int id, Trecord *rec); // repositions pair
    ~HeapBal();      // destroys H

     array_RPBal** infreq;
private:
    Thnode *freq; // space for all frequent nodes is preallocated, sqrt(u)
    int freef; // ptr to free list in freq
    Thfreq *ff; // space for all frequencies of frequent nodes prealloc idem
    int freeff; // ptr to free list in ff
    int smallest,largest; // list of frequent ones (ptrs in ff)
    // vectors for infrequent ones
    int sqrtu;
    int max;  // max freq heap used
    //Trarray *Rec; // records

    void prnH ();

};

#endif	/* _HEAP_H */

