/*  
 * Copyright (C) 2012, Andres Abeliuk, all rights reserved.
 *
 * Andres Abeliuk <aabeliuk@dcc.uchile.cl>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

int PRNC = 0;
int PRNR = 0;
int PRNP = 0;
int PRNL = 0;

int NullFreqs = 1 << (8 * sizeof (int) - 1);
int deleted_rules, trim_rules;
int tt;

#include "NPR_RPBal.h"


namespace cds_static {


 void set_field64(uint *A, int len, size_t index, size_t x) {
        if(len==0) return;
        int i=index*len/W, j=index*len-i*W;
        uint mask = ((j+len) < W ? ~0u << (j+len) : 0)
            | ((W-j) < W ? ~0u >> (W-j) : 0);
        A[i] = (A[i] & mask) | x << j;
        if (j+len>W) {
            mask = ((~0u) << (len+j-W));
            A[i+1] = (A[i+1] & mask)| x >> (W-j);
        }
    }

    uint get_field64(uint *A, size_t len,  size_t index) {
        if(len==0) return 0;
        int i=index*len/W, j=index*len-W*i;
        uint result;
        if (j+len <= W)
            result = (A[i] << (W-j-len)) >> (W-len);
        else {
            result = A[i] >> j;
            result = result | (A[i+1] << (WW-j-len)) >> (W-len);
        }
        return result;
    }



    void save_lcp(ofstream & fp, int *lcp, int len) {
        saveValue(fp, lcp, len);
    }

    int* load_lcp(ifstream & fp, int len) {
        int *lcp = loadValue<int>(fp, len);
        return lcp;
    }

    int NPR_RPBal::get_C(size_t i) const {
        return get_field64(C, b, i) - Cmin;
    }

    void NPR_RPBal::set_C(size_t i, size_t x) const {
        set_field64(C, b, i, x + Cmin);
    }

    NPR_RPBal::NPR_RPBal() {
        npr_type = RPBal_NPR;
        minsize = 0;
        factor = 0;
        rate = 0;
        trim = 0;
        len = 0;
        c = u = Cmin = len;
        alph = 0;
        C = NULL;
        Rec = NULL;
        heap = NULL;
        hash = NULL;
        L = NULL;
        rules = NULL;
    }

    NPR_RPBal::NPR_RPBal(LCP_RLCSA *lcp, RLCSA *csa, int length, size_t block_Size, int _minsize, float _factor) {
        npr_type = RPBal_NPR;
        minsize = _minsize;
        factor = _factor;
        rate = block_Size;
        trim = block_Size;
        Cmin = len = length;
        c = u = len - 1;
        Rec = new Records_Bal(factor, minsize);
        heap = new HeapBal(u, factor, minsize,Rec);
        hash = new Hash_Bal(1024*1024, factor);
        L = new Tlist [u];
        rules = new Rules(factor, minsize, length);
        Rec->assocRecords(hash, heap, L);
        alph = 0;

        printf("Length=%i\n", len);
        printf("diferential lcp\n");

        size_t last, next, val;
        b = bits(len+len - 1) + 1; // extra bit to represent negatives
        long long nb = (b * len + W - 1) / W;
        if (nb > MAXINT) {
            cout << "Memory limit excess (in NPR)" << endl;
            exit(1);
        }
	
        C = new uint[(size_t) nb];
        for (size_t i = 0; i < (size_t) nb; i++)
            C[i] = 0;
        set_C(0, len);
        last = lcp->get_LCP(1, csa);

        for (size_t i = 2; i < len; i++) {
            next = lcp->get_LCP(i, csa);
            val = len;
            val += (next-last);
            set_C(i - 1, val);
            last = next;
        }


        printf("Start Re-pair\n");
        prepare();
        printf("Finish prepare\n");
        repair();
        printf("Finish repair\n");

        printf("alph = %i\n", alph);

        uint *aux = new uint[(size_t) nb]; //for copying C without gaps
        for (size_t i = 0; i < (size_t) nb; i++)
            aux[i] = 0;
        //int* aux = new int [c];
        int i = 0;
        int j = 0;
        while (i < u) {
            set_field64(aux, b, j++, get_C(i) + len); //aux[j++] = C[i];
            i++;
            if ((i < u) && (get_C(i) < 0)) i = -get_C(i) - 1;
        }
        //memcpy(C,aux,c*sizeof(int));//C without gaps

        delete [] C;
        C = aux;
        c = j;



        compress_CV2();
        deleted_rules = trim_rule();
        sampling(rate);
	
        int maxdepth = 0;
        for (i = 0; i < n - alph; ++i) {
            if (maxdepth < rules->get_nt(i))
                maxdepth = rules->get_nt(i);
        }

        fprintf(stderr, "RePair succeeded\n\n");
        fprintf(stderr, "   Original ints: %i\n", len);
        fprintf(stderr, "   Number of rules: %i\n", n - alph);
        fprintf(stderr, "   Max  rule cover: %i\n", maxdepth);
        fprintf(stderr, "   Number of rules trimmed: %i\n", trim_rules);
        fprintf(stderr, "   Number of rules deleted: %i\n", deleted_rules);
        fprintf(stderr, "   Final sequence length: %i\n", c);
        fprintf(stderr, "   NPR Size/c: %f\n", getSize()*8.0 / len);
	
	
	

    }

    NPR_RPBal::NPR_RPBal(LCP_RLCSA *lcp, RLCSA *csa, int length, int _minsize, float _factor) {
        npr_type = RPBal_NPR;
        minsize = _minsize;
        factor = _factor;
        rate = 0;
        trim = 0;
        Cmin = len = length;
        c = u = len - 1;
        Rec = new Records_Bal(factor, minsize);
        heap = new HeapBal(u, factor, minsize, Rec);
        hash = new Hash_Bal(1024*1024, factor);
        L = new Tlist [u];
        rules = new Rules(factor, minsize, length);
        Rec->assocRecords(hash, heap, L);
        alph = 0;


        printf("Length=%i\n", len);
        printf("diferential lcp\n");

        size_t last, next, val;
        b = bits(len+len - 1) + 1; // extra bit to represent negatives
        printf("bits=%i\n", b);
        long long nb = (b * len + W - 1) / W;
        if (nb > MAXINT) {
            cout << "Memory limit excess (in NPR)" << endl;
            exit(1);
        }

        C = new uint[(size_t) nb];
        for (size_t i = 0; i < (size_t) nb; i++)
            C[i] = 0;
	
        set_C(0, len);
        last = lcp->get_LCP(1, csa);

        for (size_t i = 2; i < len; i++) {
            next = lcp->get_LCP(i, csa);
            val = len;
            val += (next-last);
            set_C(i - 1, val);
            last = next;
        }
	 printf("alph=%i \n",alph);

        printf("Start Re-pair\n");
        prepare();
        printf("Finish prepare\n");
        repair();
        printf("Finish repair\n");

        printf("alph = %i\n", alph);

        uint *aux = new uint[(size_t) nb]; //for copying C without gaps
        for (size_t i = 0; i < (size_t) nb; i++)
            aux[i] = 0;
        //int* aux = new int [c];
        int i = 0;
        int j = 0;
        while (i < u) {
            set_field64(aux, b, j++, get_C(i) + len); //aux[j++] = C[i];
            i++;
            if ((i < u) && (get_C(i) < 0)) i = -get_C(i) - 1;
        }
        //memcpy(C,aux,c*sizeof(int));//C without gaps

        delete [] C;
        C = aux;
        c = j;



        int maxdepth = 0;
        for (i = 0; i < n - alph; ++i) {
            if (maxdepth < rules->get_nt(i))
                maxdepth = rules->get_nt(i);
        }

        fprintf(stderr, "RePair succeeded\n\n");
        fprintf(stderr, "   Original ints: %i\n", len);
        fprintf(stderr, "   Number of rules: %i\n", n - alph);
        fprintf(stderr, "   Max  rule cover: %i\n", maxdepth);
        fprintf(stderr, "   Final sequence length: %i\n", c);


    }

    int NPR_RPBal::expand(int i)
 {
        int ret = 1;
        while (i >= alph && (i - alph) < trules) { 
            ret += expand(rules->get_left(i - alph));
            i = rules->get_right(i - alph);
        }
        printf("%i, ", i);

        return ret;
    }

    int NPR_RPBal::expandRule(int i, int *index, int *result)const
 {	
	int left;	
        int ret = 1;
	int r = i;
        while (r >= alph && (r - alph) < trules) {
	    left = rules->get_left(r - alph);
            ret += expandRule(left, index, result);
            r = rules->get_right(r - alph);
        }

        result[*index] = r;
	*index= *index + 1;
        return ret;
    }

    void NPR_RPBal::prepare() {
        int i, id;
        Tpair pair;
        for (i = 0; i < u; i++) {
            if (get_C(i) > alph)
                alph = get_C(i);
        }
        n = ++alph;
        printf("alph=%i\n", alph);
        for (i = 0; i < c - 1; i++) {
            pair.left = get_C(i);
            pair.right = get_C(i + 1);
            id = hash->searchHash(pair, Rec->records);
            if (id == -1) // new pair, insert
            {
                id = Rec->insertRecordDiff(pair, len);
                L[i].next = -1;
            } else {
                L[i].next = Rec->records[id].cpos;
                L[L[i].next].prev = i;
                heap->incFreq(id, Rec->records);
            }
            L[i].prev = -id - 1;
            Rec->records[id].cpos = i;
            if (PRNL && (i % 10000 == 0)) printf("Processed %i chars\n", i);
        }
        heap->purgeHeap(Rec);
    }

    void prnSym(int c) {
        printf("%i", c);
    }

    void NPR_RPBal::prnC()
 {
        int i = 0;
        printf("C[1..%i] = ", c);
        while (i < u) {
            prnSym(get_C(i));
            printf(" ");
            i++;
            if ((i < u) && (get_C(i) < 0)) i = -get_C(i) - 1;
        }
        printf("\n\n");
        // printf ("Physical C[1..%i] = ",u);
        // for (i=0;i<u;i++)
        //       { prnSym(C[i]);
        //  	printf (" ");
        //       }
        //     printf ("\n\n");
    }

    void NPR_RPBal::prnRec()
 {
        int i;
        printf("Active pairs:\n");
        for (i = 0; i < Rec->size; i++) {
            printf("\t(");
            prnSym(Rec->records[i].pair.left);
            printf(",");
            prnSym(Rec->records[i].pair.right);
            printf("), %i occs", Rec->records[i].freq);
            printf(" cover: %i \n", Rec->records[i].nt);
        }
        printf("\n");
    }

    // to avoid another list to access the sparse C we thread it using the
    // empty space. if next cell of an active cell is negative, it is
    // (minus) a ptr to the next occ. idem previous cell to previous occ,
    // except that next ptr dominates over prev ptr if they must be in
    // the same cell. but in this case one can find prev in O(1) anyway.

    int NPR_RPBal::repair()
 {
        int oid, id, cpos, min, lminpos, rminpos;
        Trecord *rec, *orec;
        Tpair pair;
        // if (fwrite(&alph,sizeof(int),1,R) != 1) return -1;
        if (PRNC) prnC();
        while (n + 1 > 0) {
            if (PRNR) prnRec();
            oid = heap->extractMax();
            if (oid == -1) break; // the end!!
            orec = &Rec->records[oid];
            cpos = orec->cpos;
            //if (fwrite (&orec->pair,sizeof(Tpair),1,R) != 1) return -1;
            //save rule
            rules->insertRule(orec->pair.left, orec->pair.right, orec->min,orec->minleftpos, orec->minrightpos, orec->nt, orec->sum);
	   
            if (PRNP) {
                printf("Chosen pair %i = (", n);
                prnSym(orec->pair.left);
                printf(",");
                prnSym(orec->pair.right);
                printf(") (%i occs)\n", orec->freq);
                printf("min : %i\n", orec->min);
		printf("min pos : %i\n", orec->minleftpos);
                printf("terminals : %i\n", orec->nt);
                printf("Sum : %i\n", orec->sum);
            }
            while (cpos != -1) {
                int ant, sgte, ssgte;
                // replacing bc->e in abcd, b = cpos, c = sgte, d = ssgte
                if (get_C(cpos + 1) < 0) sgte = -get_C(cpos + 1) - 1;
                else sgte = cpos + 1;
                if ((sgte + 1 < u) && (get_C(sgte + 1) < 0)) ssgte = -get_C(sgte + 1) - 1;
                else ssgte = sgte + 1;
                // remove bc from L
                if (L[cpos].next != -1)
                    L[L[cpos].next].prev = -oid - 1;
                orec->cpos = L[cpos].next;
                if (ssgte != u) // there is ssgte
                { // remove occ of cd
                    pair.left = get_C(sgte);
                    pair.right = get_C(ssgte);
                    id = hash->searchHash(pair, Rec->records);
                    if (id != -1) // may not exist if purgeHeap'd
                    {
                        if (id != oid) heap->decFreq(id, Rec); // not to my pair!
                        if (L[sgte].prev != NullFreqs) //still exists(not removed)
                        {
                            rec = &Rec->records[id];
                            if (L[sgte].prev < 0) // this cd is head of its list
                                rec->cpos = L[sgte].next;
                            else L[L[sgte].prev].next = L[sgte].next;
                            if (L[sgte].next != -1) // not tail of its list
                                L[L[sgte].next].prev = L[sgte].prev;
                        }
                    }
                    // create occ of ed
                    pair.left = n;
                    //find min
                    int rule = pair.right;
                    int rmin, nnt, sum, lpos, rpos;
                    rmin = rules->get_sum(n - alph);
                    nnt = rules->get_nt(n - alph);
                    sum = rules->get_sum(n - alph);
 
                    if (rule >= alph) {//not terminal
                        rmin += rules->get_min(rule - alph); // add left block
                        nnt += rules->get_nt(rule - alph);
                        sum += rules->get_sum(rule - alph);
			lpos = rules->get_lminpos(rule - alph);
			rpos = rules->get_rminpos(rule - alph);
                    } else {
                        rmin += (rule - len);
                        nnt += 1;
                        sum += (rule - len);
			lpos = rpos = 1;
                    }


                    if (rules->get_min(n - alph) < rmin){ //min is the same as the left rule
                        min = rules->get_min(n - alph);
			lminpos = rules->get_lminpos(n - alph);
			rminpos = rules->get_rminpos(n - alph);
                    }else if(rules->get_min(n - alph) > rmin){ 
			min = rmin;
			lminpos = rules->get_nt(n - alph) + lpos;
			rminpos = rules->get_nt(n - alph) + rpos;
		    }else{
			min = rules->get_min(n - alph);
			lminpos = rules->get_lminpos(n - alph);
			rminpos = rules->get_nt(n - alph) + rpos;
		    }


                    id = hash->searchHash(pair, Rec->records);
                    if (id == -1) // new pair, insert
                    {
                        id = Rec->insertRecordMin(pair, min, lminpos, rminpos, nnt, sum);
                        rec = &Rec->records[id];
                        L[cpos].next = -1;
                    } else {
                        heap->incFreq(id, Rec->records);
                        rec = &Rec->records[id];
                        L[cpos].next = rec->cpos;
                        L[L[cpos].next].prev = cpos;
                    }
                    L[cpos].prev = -id - 1;
                    rec->cpos = cpos;
                }
                if (cpos != 0) // there is ant
                { // remove occ of ab
                    if (get_C(cpos - 1) < 0) {
                        ant = -get_C(cpos - 1) - 1;
                        if (ant == cpos) // sgte and ant clashed -> 1 hole
                            ant = cpos - 2;
                    } else ant = cpos - 1;
                    pair.left = get_C(ant);
                    pair.right = get_C(cpos);
                    id = hash->searchHash(pair, Rec->records);
                    if (id != -1) // may not exist if purgeHeap'd
                    {
                        if (id != oid) heap->decFreq(id, Rec); // not to my pair!
                        if (L[ant].prev != NullFreqs) //still exists (not removed)
                        {
                            rec = &Rec->records[id];
                            if (L[ant].prev < 0) // this ab is head of its list
                                rec->cpos = L[ant].next;
                            else L[L[ant].prev].next = L[ant].next;
                            if (L[ant].next != -1) // it is not tail of its list
                                L[L[ant].next].prev = L[ant].prev;
                        }
                    }
                    // create occ of ae
                    pair.right = n;
                    //find min
                    int rule = pair.left;
                    int rmin, lmin, nnt, sum, lpos, rpos, lnt;

                    if (rule >= alph) {//not terminal
                        rmin = rules->get_sum(rule - alph) + rules->get_min(n - alph);
                        lmin = rules->get_min(rule - alph);
			lpos = rules->get_lminpos(rule - alph);
			rpos = rules->get_rminpos(rule - alph);
			lnt = rules->get_nt(rule - alph);
                        nnt = lnt + rules->get_nt(n - alph);
                        sum = rules->get_sum(rule - alph) + rules->get_sum(n - alph);
                    } else {
                        rmin = rule - len + rules->get_min(n - alph);
                        lmin = rule - len;
			lpos = rpos = 1;
			lnt = 1;
                        nnt = lnt + rules->get_nt(n - alph);
                        sum = rules->get_sum(n - alph) + rule - len;
                    }

                    if (lmin < rmin){//min at left rule
                        min = lmin;
			lminpos = lpos;
			rminpos = rpos;
                    }else if(lmin > rmin){
			 min = rmin;
		    	 lminpos = lnt + rules->get_lminpos(n - alph);
			 rminpos = lnt + rules->get_rminpos(n - alph);
		    }else{
		    	min = lmin;
			lminpos = lpos;
			rminpos = lnt + rules->get_rminpos(n - alph);
		    }

                    id = hash->searchHash(pair, Rec->records);
                    if (id == -1) // new pair, insert
                    {
                        id = Rec->insertRecordMin(pair, min, lminpos, rminpos, nnt, sum);
                        rec = &Rec->records[id];
                        L[ant].next = -1;
                    } else {
                        heap->incFreq(id, Rec->records);
                        rec = &Rec->records[id];
                        L[ant].next = rec->cpos;
                        L[L[ant].next].prev = ant;
                    }
                    L[ant].prev = -id - 1;
                    rec->cpos = ant;
                }
                set_C(cpos, n); //C[cpos] = n;
                if (ssgte != u) set_C(ssgte - 1, -cpos - 1); //C[ssgte-1] = -cpos-1;
                set_C(cpos + 1, -ssgte - 1); //C[cpos+1] = -ssgte-1;
                c--;
                orec = &Rec->records[oid]; // just in case of Rec.records realloc'd
                cpos = orec->cpos;
            }
            if (PRNC) prnC();
            Rec->removeRecord(oid);
          
            n++;
            heap->purgeHeap(Rec); // remove freq 1 from heap
            if (c < factor * u) // compact C
            {
                int i, ni;
                i = 0;
                for (ni = 0; ni < c - 1; ni++) {
                    set_C(ni, get_C(i));
                    L[ni] = L[i];
                    if (L[ni].prev < 0) {
                        if (L[ni].prev != NullFreqs) // real ptr
                            Rec->records[-L[ni].prev - 1].cpos = ni;
                    } else L[L[ni].prev].next = ni;
                    if (L[ni].next != -1) L[L[ni].next].prev = ni;
                    i++;
                    if (get_C(i) < 0) i = -get_C(i) - 1;
                }

                set_C(ni, get_C(i));
                u = c;
                //realloc
                // C = (int*)realloc (C, c * sizeof(int));
                size_t nb = (b * c + W - 1) / W;
                uint *aux1 = new uint[(size_t) nb];
                for (size_t i = 0; i < (size_t) nb; i++)
                    aux1[i] = 0;

                for (i = 0; i < c; ++i)
                    set_field64(aux1, b, i, get_C(i) + len);
                delete [] C;
                C = aux1;

                //L = (Tlist*)realloc (L, c * sizeof(Tlist));
                Tlist *aux2 = new Tlist[c];
                for (i = 0; i < c; ++i)
                    aux2[i] = L[i];
                delete [] L;
                L = aux2;

                Rec->assocRecords(hash, heap, L);
            }
        }
        return 0;
    }

    void NPR_RPBal::get_min_pos(RLCSA *csa, LCP_RLCSA *lcp){
        int nt, min, cval, first, lcp_pos, lcp_value, max_pos;
        size_t *minpos = new size_t[c];

        first = max_pos = 0;

        for (size_t i = 0; i < c; ++i) {
            cval = get_C(i);
            //check is it rule or terminal
            if (cval >= alph) { //rule
                nt = rules->get_nt(cval - alph);
                min = rules->get_min(cval - alph);
            } else { //terminal
                nt = 1;
                min = cval - len;
            }
            NSV_rule(&lcp_pos, &lcp_value, min + 1, cval, first, lcp->get_LCP(first, csa), csa, lcp);
            minpos[i] = lcp_pos;
            first += nt;
            if(max_pos<lcp_pos) max_pos = lcp_pos;
        }

        bmin = bits(max_pos - 1);
        long long nb = (b * c + W - 1) / W;

        min_pos = new uint[(size_t) nb];
        for (size_t i = 0; i < (size_t) nb; i++)
            min_pos[i] = 0;
        
        for (size_t i = 0; i < c; ++i)
             set_field64(min_pos, bmin, i, minpos[i]);
    }

   
    void NPR_RPBal::compress_CV2() {
        int j, nt1, min1, sum1, lpos1, rpos1, nt2, min2, sum2, lpos2, rpos2, cval1, cval2, rule_nt, rule_sum, rule_min, rule_lminpos, rule_rminpos, c_min, run_len, nrule;
        size_t *CC = new size_t[c];
        C_rules = new Rules(factor, minsize, len);

        rule_nt = rule_sum = rule_min = run_len = 0;
        c_min = nrule = n;

        j = 0;
        for (int i = 0; i < c - 1; ++i) {
            cval1 = get_C(i);
            cval2 = get_C(i + 1);
            if (c_min > cval1 - len)
                c_min = cval1 - len;

            //check is it rule or terminal
            if (cval1 >= alph) { //rule
                nt1 = rules->get_nt(cval1 - alph);
                min1 = rules->get_min(cval1 - alph);
		lpos1 = rules->get_lminpos(cval1 - alph);
		rpos1 = rules->get_rminpos(cval1 - alph);
                sum1 = rules->get_sum(cval1 - alph);
            } else { //terminal
                nt1 = 1;
		lpos1 = rpos1 = 1;
                min1 = cval1 - len;
                sum1 = min1;
            }
            //check is it rule or terminal
            if (cval2 >= alph) { //rule
                nt2 = rules->get_nt(cval2 - alph);
                min2 = rules->get_min(cval2 - alph);
		lpos2 = rules->get_lminpos(cval2 - alph);
		rpos2 = rules->get_rminpos(cval2 - alph);
                sum2 = rules->get_sum(cval2 - alph);
            } else { //terminal
                nt2 = 1;
		lpos2 = rpos2 = 1;
                min2 = cval2 - len;
                sum2 = min2;
            }

            if (nt1 > trim) {
                CC[j++] = cval1;
                run_len = 0;
                rule_nt = 0;
            } else {
                if (run_len == 0) {
                    if (nt1 + nt2 <= trim) {
                        rule_nt = nt1 + nt2;
                        rule_sum = sum1 + sum2;
                        if (min1 < sum1 + min2){
                            rule_min = min1;
			    rule_lminpos = lpos1;
			    rule_rminpos = rpos1;
                        }else if(min1 < sum1 + min2){
			    rule_min = sum1 + min2;
			    rule_lminpos = nt1 + lpos2;
			    rule_rminpos = nt1 + rpos2;
			}else{
			    rule_min = min1;
			    rule_lminpos = lpos1;
			    rule_rminpos = nt1 + rpos2;
			}
                        ++run_len;
                    } else {
                        CC[j++] = cval1;
                        run_len = 0;
                        rule_nt = 0;
                    }
                } else {//nt_sum>0 && run_length>0
                    if (rule_nt + nt2 <= trim) {
                        rule_nt += nt2;
                        if (rule_sum + min2 < rule_min){//min in right node
                            rule_min = rule_sum + min2;
			    rule_lminpos = rule_nt + lpos2;
			    rule_rminpos = rule_nt + rpos2;
			}
                        rule_sum += sum2;
                        ++run_len;
                    } else {
                        CC[j++] = nrule++;
                        C_rules->insertTrimRule(rule_min, rule_nt, rule_lminpos, rule_rminpos, rule_sum);
                        run_len = 0;
                        rule_nt = 0;
                    }
                }
            }
        }
        if (run_len > 0) {
            CC[j++] = nrule++;
            C_rules->insertTrimRule(rule_min, rule_nt, rule_lminpos, rule_rminpos, rule_sum);
        } else if (run_len == 0)
            C[j++] = cval2;

        c_min = -c_min + 1;
        size_t nb = bits(nrule - len + c_min - 1);
        printf("nb = %i\n", nb);
        long long s = (nb * j + W - 1) / W;
        uint *aux = new uint[(size_t) s]; //for copying C without gaps
        for (size_t i = 0; i < (size_t) s; i++)
            aux[i] = 0;


        for (int i = 0; i < j; ++i) {
            set_field64(aux, nb, i, CC[i] - len + c_min);
        }
        printf("size of NEW rules=%i\n", nrule - len);
        delete [] C;
        delete [] CC;

        C = aux;
        b = nb;
        Cmin = c_min;
        n = n - len;
        alph = alph - len;
   //     printf("New alph=%i\n", alph);
   //     printf("rules=%i\n", n - alph);
   //     printf("New n=%i\n", n);

        //rename rules
        int *Rleft, *Rright;
        int nval, nmin;
        Rleft = new int[n - alph];
        Rright = new int[n - alph];
        nmin = 0;
        for (int i = 0; i < n - alph; ++i) {
            nval = rules->get_left(i);
            nval -= len;

            Rleft[i] = nval;
            if (nval < nmin)
                nmin = nval;

            nval = rules->get_right(i);
            nval -= len;
            Rright[i] = nval;
            if (nval < nmin)
                nmin = nval;
        }
     //   printf("nmin=%i\n", nmin);


        nmin = -nmin;
        rules->set_Rmin(nmin);
        for (int i = 0; i < n - alph; ++i) {
            nval = Rleft[i] + nmin;
            rules->set_left(i, nval);

            nval = Rright[i] + nmin;
            rules->set_right(i, nval);
        }
        delete [] Rleft;
        delete [] Rright;

        //printf("Rules from text errased = %i of %i in rules=%i\n", c-j,c,nr);
        c = j;

    }
	

    // returns position in compressed text containing value at position i
    // and position of first terminal of rule

    void NPR_RPBal::get_pos_seq(int i, int *pos, int *first, RLCSA *csa, LCP_RLCSA *lcp)const {
        int p, sum;
        p = *pos;
        sum = *first;

        while (i > sum) {
            //check is it rule or terminal
            if (get_C(p) >= alph) //rule
                sum += rules->get_nt(get_C(p) - alph);
            else //terminal
                sum += 1;
            ++p;
        }
        p = p - 1;
        *pos = p;
        if (get_C(p) >= alph) //rule
            *first = sum - rules->get_nt(get_C(p) - alph);
        else
            *first = sum - 1;
    }

    //binary search in Array F , containing sample of LCP positions

    int NPR_RPBal::BinarySearch(int i, int low, int high, int last) const {
        if (high < low)
            return last; // not found
        int mid = low + (high - low) / 2;
        if (F[mid] > i)
            return BinarySearch(i, low, mid - 1, mid);

        else if (F[mid] < i) {
            if (F[mid + 1] > i)
                return mid;
            else
                return BinarySearch(i, mid + 1, high, mid);
        }
        else
            return mid; // found
    }

    // returns position in compressed text containing value at position i in LCP
    // and position of first terminal of rule

    void NPR_RPBal::get_pos(int i, int *pos, int *first, RLCSA *csa, LCP_RLCSA *lcp)const {
        int high;

        if (i < F[0]) {
            get_pos_seq(i, pos, first, csa, lcp);
        } else {
            high = floor((c - 1) / rate);
            *pos = BinarySearch(i, 0, high - 1, -1);
            *first = F[*pos];
            *pos = (*pos + 1) * rate;
            get_pos_seq(i, pos, first, csa, lcp);
        }
    }


    //creates sampling of first positions of the LCP in compressed text
    // bynary search over sampling gives rule where position is

    void NPR_RPBal::sampling(int rate) {

        int i, sum, size;
        sum = 0;
        size = floor((c - 1) / rate);
        F = new int[size];
        // A = new int[size];

        for (i = 1; i <= c; ++i) {
            if (get_C(i - 1) >= alph) //rule
                sum += rules->get_nt(get_C(i - 1) - alph);
            else //terminal
                sum += 1;

            //write sample
            if (i % rate == 0) {
                F[i / rate - 1] = sum;
                // A[i/rate - 1] = lcp->get_LCP(i,csa);
            }

        }

    }

    int NPR_RPBal::LCP_seq(int first, int *lcp_pos, int *lcp_value, int value, int cover, RLCSA *csa, LCP_RLCSA *lcp)const {
        //can asume min < value is in this tree
        int i, v;
        for (i = first + 1; i <= first + cover; ++i){
            if ((v = lcp->get_LCP(i, csa)) < value) { 
                *lcp_value = v;
                *lcp_pos = i;
                return 1;
            }
	}
        return 0;

    }

    int NPR_RPBal::LCP_seq_left(int first, int *lcp_pos, int *lcp_value, int value, int cover, RLCSA *csa, LCP_RLCSA *lcp)const {
        //can asume min < value is in this tree
        int i, v;
        for (i = first; i > first - cover; --i) { 
            if ((v = lcp->get_LCP(i, csa)) < value) {
                *lcp_value = v;
                *lcp_pos = i;
                return 1;
            }
        }
        return 0;

    }

    int NPR_RPBal::trim_rule() {
        HashMap *deleted = new HashMap();
        HashMap *trimed = new HashMap();
        int i, left, right, cval;
        trim_rules = 0;

        for (i = 0; i < n - alph; ++i) {
            //add rules < trim
            if (rules->get_nt(i) <= trim) {
                if (deleted->get(i) == -1) {
                    deleted->put(i);
                }
            }

        }


        //cant delete if rule is in C
        for (i = 0; i < c; ++i) {
            cval = get_C(i);
            if (cval >= alph) {//is rule
                if (deleted->get(cval - alph) != -1) {
                    deleted->erase(cval - alph);
                    if (trimed->get(cval - alph) == -1) {
                        trimed->put(cval - alph);
                        ++trim_rules;
                    }
                }
            }
        }


        for (i = 0; i < n - alph; ++i) {
            if (rules->get_nt(i) > trim) {
                //check sons
                left = rules->get_left(i);
                right = rules->get_right(i);
                if (left >= alph) { //rule
                    if (deleted->get(left - alph) != -1) {//if(rules->get_nt(left-alph) <= trim){
                        // if(deleted->get(left-alph) != -1)
                        deleted->erase(left - alph);
                        if (trimed->get(left - alph) == -1) {
                            trimed->put(left - alph);
                            ++trim_rules;
                        }
                    }
                }
                if (right >= alph) { //rule
                    if (deleted->get(right - alph) != -1) {//if(rules->get_nt(right-alph) <= trim){
                        //if(deleted->get(right-alph) != -1)
                        deleted->erase(right - alph);
                        if (trimed->get(right - alph) == -1) {
                            trimed->put(right - alph);
                            ++trim_rules;
                        }
                    }
                }

            }
        }
	
	printf("Rules deleted = %i\n", deleted->used);
	printf("Rules trimmed = %i\n", trimed->used);
        n -= deleted->used;
        n -= trimed->used;

        printf("Rules= %i\n", n - alph);
        int *location;
        //delete and reallocate rules
        location = rules->trim(deleted, trimed, C_rules, alph);





        //add C rules
        size_t value;
        //rename rules in C
        for (i = 0; i < c; ++i) {
            //rule or terminal
            if (get_C(i) >= alph) { //rule
                value = location[get_C(i) - alph];

                value += alph;
                set_C(i, value);
            }
        }

        return deleted->used;

    }

    int NPR_RPBal::trim_tree()const {
        int i, t;
        t = 0;
        for (i = 0; i < n - alph; ++i)
            if (rules->get_nt(i) <= trim) {
                t++;
                // Rules[i].min = 0;
                // Rules[i].sum = 0;
                // delete [] &Rules[i].pair;
                // Rules[i].pair = NULL;
            }
        return t;
    }

    int NPR_RPBal::NSV_rule(int *lcp_pos, int *lcp_value, int value, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const {
        //can asume min < value is in this tree
        int left, right, cover, min, minpos, lcpvalue, sum, rlcpvalue, rfirst;
	//printf("NSV_RULE value=%i, rule=%i, first=%i, cover=%i,\n",value,rule,first,cover);
	if(first_lcp < value){		
	    *lcp_pos = first;
            *lcp_value = first_lcp;
            return 1;
        }	

        if (rule < alph) {//terminal 
            *lcp_pos = first + 1;
            *lcp_value = first_lcp + rule;
            return 1;
        }
	
	
        cover = rules->get_nt(rule - alph);
	min = rules->get_min(rule - alph) + first_lcp;
	if(value == min+1){//we are sure min of rule is nsv, no need of scan LCP (if not value>x>min could be to to the left of min)
		minpos = rules->get_lminpos(rule - alph);
		*lcp_pos = first + minpos;
	    	*lcp_value = first_lcp + min;
	//printf("NSV minPos lcp_pos =%i, lcp_value=%i, first =%i, minpos =%i, covervalue=%i, REALvalue=%i  \n",*lcp_pos,*lcp_value, first,minpos, cover, lcp->get_LCP(*lcp_pos, csa));  
	     	return 1;
	}

	//min = rules->get_min(rule - alph);
	//min += lcp->get_LCP(first, csa);  
        if (rule - alph >= trules) { //no children
	//printf("LCP_seq first=%i, cover=%i, min=%i\n",first,cover, min);
            LCP_seq(first, lcp_pos, lcp_value, value, cover, csa, lcp);
            return 1;
        }
	
        left = rules->get_left(rule - alph);
        right = rules->get_right(rule - alph);

        lcpvalue = first_lcp;
        if (left >= alph) { //rule
            min = rules->get_min(left - alph) + lcpvalue;
            cover = rules->get_nt(left - alph);
	    sum = rules->get_sum(left - alph);
        } else { //terminal
            min = lcpvalue;
            cover = 1;
	    sum = left;
        }
	rfirst = first + cover;
	rlcpvalue = lcpvalue + sum;
	
        if (min < value)
           return NSV_rule(lcp_pos, lcp_value, value, left, first, lcpvalue, csa, lcp);
	else if (rlcpvalue < value) {//first not included in rule
	    *lcp_pos = rfirst;
	    *lcp_value = rlcpvalue;
	    return 1;
	}else {// min is on right rule
            return NSV_rule(lcp_pos, lcp_value, value, right, rfirst, rlcpvalue, csa, lcp);
        }

    }

    int NPR_RPBal::first_NSV(int *lcp_pos, int *lcp_value, int i, int value, int rule, int first, int first_lcp,  RLCSA *csa, LCP_RLCSA *lcp)const {
        int left, right, lmin, rmin, rpos, rfirst, sum, cover, val;
	
        if (rule < 0) //terminal
            cover = 1;
        else
            cover = rules->get_nt(rule);
        
	rpos = first + cover;

	//printf("first_NSV i=%i, value=%i, pos=%i, rule=%i, first=%i ,first_lcp=%i, cover=%i \n",i,value,*lcp_pos,rule,first,first_lcp, cover);  
        if (rpos == i) // i is in the rightmost position of the tree
            return 0;
	
//	if (i <= first){ // i is to the right of the rule
//	    return  NSV_rule(lcp_pos, lcp_value, value, rule+alph, first, first_lcp, csa, lcp);
//	}
        if (rule >= trules) {// artifial C-rule, no sons
            if (LCP_seq(i, lcp_pos, lcp_value, value, cover, csa, lcp))
                return 1;
            else return 0;
        }

        left = rules->get_left(rule);
        right = rules->get_right(rule);
  
        val = first_lcp;
        if (left >= alph) { //rule
            rpos = first + rules->get_nt(left - alph) - 1; //cout << "nt "<< rules->get_nt(left - alph)  <<  endl;//rightmost position of terminal in left rule
            lmin = rules->get_min(left - alph) + val;   //cout << "min "<< rules->get_min(left - alph)  <<  endl;
	    sum = rules->get_sum(left - alph);	//cout << "sum "<< sum  <<  endl;
        } else { //terminal
            rpos = first;
            lmin = val;
	    sum = left;
        }
        rfirst = rpos + 1; //first position of right rule
	
        val += sum; 
        if (right >= alph) {//rule
            rmin = rules->get_min(right - alph) + val;
        } else
            rmin = val;
		
	//printf("lmin=%i, rmin=%i ,rpos=%i, lcpvalue[%i]=%i, lcp_real=%i \n",lmin,rmin,rpos, rfirst, val, lcp->get_LCP(rfirst, csa));
        if (lmin < value) {
            if (rpos >= i) {//i is en left rule
                if (first_NSV(lcp_pos, lcp_value, i, value, left - alph, first, first_lcp, csa, lcp))
                    return 1;
		if (val < value) {//first not included in rule
		    *lcp_pos = rfirst;
		    *lcp_value = val;
		    return 1;
		}
                else {// check right rule, i is in left node
                    if (rmin < value) {
                        NSV_rule(lcp_pos, lcp_value, value, right, rfirst, val, csa, lcp);
                        return 1;
                    } else
                        return 0; //end
                }
            } else { // i is in right node
                if (rmin < value)
                    return first_NSV(lcp_pos, lcp_value, i, value, right - alph, rfirst, val, csa, lcp);
                else
                    return 0;
            }
        }else if (val < value && rfirst > i) {//first not included in rule
	    *lcp_pos = rfirst;
	    *lcp_value = val;
	    return 1;
	}else if (rmin < value) {
            if (rpos < i) { // i is in right node
                return first_NSV(lcp_pos, lcp_value, i, value, right - alph, rfirst, val, csa, lcp);
            } else {
               return NSV_rule(lcp_pos, lcp_value, value, right, rfirst, val, csa, lcp);
               //return 1;
            }
        }else{
	    return 0;
	}
    }






 
    // Return the position of the next smaller value of i (NSV[i])

    size_t NPR_RPBal::find_NSV(int i, int *poslcp, int *lcp_value, RLCSA *csa, LCP_RLCSA *lcp) const {
        int value, pos, first, min, k, lcpvalue, rule;
	  
        if (i == 0)
            return len;
	
	
        pos = first = 0;
        value = lcp->get_LCP(i, csa);
        if (value == 0)
            return len;
	
       get_pos(i, &pos, &first, csa, lcp);

     //  printf("Find_NSV i=%i, value=%i, pos=%i, first=%i\n",i,value,pos,first);
      

        //first rule of compressed text
        lcpvalue = lcp->get_LCP(first, csa); 
        rule = get_C(pos);
	
        if (rule >= alph) {//rule
            min = rules->get_min(rule - alph) + lcpvalue; 
            if (min < value) { 
                if (first_NSV(poslcp, lcp_value, i, value, rule - alph, first, lcpvalue, csa, lcp))
                    return *poslcp;
            }
            first += rules->get_nt(rule - alph);
	    lcpvalue += rules->get_sum(rule - alph);
        } else{
            ++first;
	    lcpvalue += rule;
	}
	
	 //first is not included in rule
        //lcpvalue = lcp->get_LCP(first, csa);
        for (k = pos + 1; k < c; ++k) { 
	    if(lcpvalue < value){
		*lcp_value = lcpvalue;
                *poslcp = first;
            	return *poslcp;
	    }
            rule = get_C(k);
            if (rule >= alph) {//rule 
                min = rules->get_min(rule - alph) + lcpvalue; 
                if (min < value) {
                    NSV_rule(poslcp, lcp_value, value, rule, first, lcpvalue, csa, lcp);
                    return *poslcp;
                }
                first += rules->get_nt(rule - alph);
                lcpvalue += rules->get_sum(rule - alph);
            } else {
                min = rule + lcpvalue;
                if (min < value) {  
                    *lcp_value = min;
                    *poslcp = first+1;
                    return *poslcp;
                }
                ++first;
                lcpvalue += rule;
            }

        }
        
        return len;

    }
  

    size_t NPR_RPBal::find_NSV(size_t i, RLCSA *csa, LCP_RLCSA *lcp) const {
        int poslcp, lcp_value, nsv;
        nsv = find_NSV(i, &poslcp, &lcp_value, csa, lcp);
        return nsv;

    }

    int NPR_RPBal::PSV_rule(int *lcp_pos, int *lcp_value, int value, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const {
        //can asume min < value is in this tree
        int left, right, cover, min, rfirst, val, sum, lcpval, minpos;
        //printf("PSV_rule value=%i, lcp_pos=%i, lcp_value=%i,rule=%i, first=%i, first_lcp=%i alph=%i \n",value,*lcp_pos,*lcp_value, rule,first,first_lcp,alph);      
	
	if (rule < alph) {//terminal
            *lcp_pos = first;
            *lcp_value = first_lcp;
            return 1;
        }

        cover = rules->get_nt(rule - alph);
	min = rules->get_min(rule - alph) + first_lcp;
	if(value == min+1){//we are sure min of rule is nsv, no need of scan LCP (if not value>x>min could be to to the left of min)
		minpos = rules->get_rminpos(rule - alph);
		*lcp_pos = first + minpos;
	    	*lcp_value = first_lcp + min;
	//printf("NSV minPos lcp_pos =%i, lcp_value=%i, first =%i, minpos =%i, covervalue=%i, REALvalue=%i  \n",*lcp_pos,*lcp_value, first,minpos, cover, lcp->get_LCP(*lcp_pos, csa));  
	     	return 1;
	}

	
        if (rule - alph >= trules) {//trim rule, no sons
            if (LCP_seq_left(first + cover, lcp_pos, lcp_value, value, cover, csa, lcp))
                return 1;
            else return 0;
        }

        left = rules->get_left(rule - alph);
        right = rules->get_right(rule - alph);

        if (left >= alph){ //rule
            rfirst = first + rules->get_nt(left - alph);
	    sum = rules->get_sum(left - alph);
	}
        else{ //terminal
            rfirst = first + 1;
	    sum = left;
	}

        val = first_lcp + sum;
        if (right >= alph) { //rule
            min = rules->get_min(right - alph) + val; 
            cover = rules->get_nt(right - alph);
	    sum = rules->get_sum(right - alph);
        } else { //terminal
            min = right + val;
            cover = 1;
	    sum = right;
        }
	lcpval = val + sum;
	if (lcpval < value) {//first not included in rule
            *lcp_pos = rfirst + cover;
            *lcp_value = lcpval;
            return 1;
        }

	//printf("PSV_rule left=%i, right=%i,rfirst=%i, min=%i, val=%i, get_min=%i\n", left , right, rfirst, min, val, rules->get_min(right - alph));

        if (min < value)
            return PSV_rule(lcp_pos, lcp_value, value, right, rfirst, val, csa, lcp);
	else if (val < value) {//first not included in rule
	    *lcp_pos = rfirst;
	    *lcp_value = val;
	    return 1;
	}else {// min is on left rule
            return PSV_rule(lcp_pos, lcp_value, value, left, first, first_lcp, csa, lcp);
        }

    }

    int NPR_RPBal::first_PSV(int *pos, int *lcp_value, int i, int value, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const {
        int left, right, lmin, rmin, rfirst, cover, sum, lcpvalue, rlcpvalue, rlast;

        //rpos = first + Rules[rule-alph].nt - 1;
        if (first >= i) // i is in the leftmost position of the rule
            return 0;
	
        if (rule < 0) //terminal
            cover = 1;
        else
            cover = rules->get_nt(rule);
	
	//printf("first_PSV i=%i, value=%i, pos=%i, rule=%i, first=%i , cover=%i \n",i,value,*pos,rule,first, cover);  

	if (first+cover <= i){ // i is to the right of the rule
	    return  PSV_rule(pos, lcp_value, value, rule+alph, first, first_lcp, csa, lcp);
	}
        if (cover <= trim) { //no children, scan sequentially//
            if (LCP_seq_left(i, pos, lcp_value, value, cover, csa, lcp))
                return 1;
            else return 0;
        }

        if (rule >= trules) {// artifial C-rule, no sons
            if (LCP_seq_left(i, pos, lcp_value, value, cover, csa, lcp))
                return 1;
            else return 0;
        }
				
        left = rules->get_left(rule);
        right = rules->get_right(rule);

	lcpvalue = first_lcp;	 
        if (left >= alph) { //rule
            rfirst = first + rules->get_nt(left - alph); //first position of right rule	
            lmin = rules->get_min(left - alph) + lcpvalue;
	    sum = rules->get_sum(left - alph);
        } else { //terminal
            rfirst = first+1;
	    lmin = lcpvalue;
	    sum = left;
        }	
	lcpvalue += sum;

        if (right >= alph) {//rule
            rmin = rules->get_min(right - alph) + lcpvalue;
	    sum = rules->get_sum(right - alph);
	    cover = rules->get_nt(right - alph);
        } else{
            rmin = lcpvalue;
	    sum = right;
	    cover = 1;
	}
	rlcpvalue = lcpvalue + sum; 
	rlast = rfirst + cover;
	if (rlcpvalue < value && rlast<i) {//first not included in rule
            *pos = rlast;
            *lcp_value = rlcpvalue;
            return 1;
        }

       //printf("lmin=%i, rmin=%i, lcpvalue[%i]=%i, left=%i, right=%i\n",lmin,rmin, rfirst, lcpvalue, left, right );
	
	if (rmin < value) { 
	    if (first_PSV(pos, lcp_value, i, value, right - alph, rfirst, lcpvalue, csa, lcp)>0) 
		return 1;
            else {// check left rule
		if (lcpvalue < value && rfirst<i ) {//first not included in rule
		    *pos = rfirst;
		    *lcp_value = lcpvalue;
		    return 1;
		}
                if (lmin < value) {
                   return first_PSV(pos, lcp_value, i, value, left - alph, first, first_lcp, csa, lcp);
                } else  
			return 0; //end
            }
	}else if (lcpvalue < value && rfirst<i ) {//first not included in rule
	    *pos = rfirst;
	    *lcp_value = lcpvalue;
	    return 1;
	}else if (lmin < value) {
           return first_PSV(pos, lcp_value, i, value, left - alph, first, first_lcp, csa, lcp);
	} else  
		return 0; //end

    }


    // Return the position of PSV[i]+1

    size_t NPR_RPBal::find_PSV(int i, int *poslcp, int *lcp_value, RLCSA *csa, LCP_RLCSA *lcp) const {
        int value, pos, first, min, k, lcpvalue, cvalue;

        pos = first = 0;
        value = lcp->get_LCP(i, csa);
        if (value == 0)
            return 0;

        get_pos(i, &pos, &first, csa, lcp);
        //printf("find_PSV i=%i, value=%i, pos=%i, first=%i, C[pos]=%i\n",i,value,pos,first,get_C(pos));
		  
        //first rule of compressed text
	lcpvalue = lcp->get_LCP(first, csa);
        cvalue = get_C(pos);
        if (cvalue >= alph) {//rule
            min = rules->get_min(cvalue - alph) + lcpvalue;
            if (min < value) {
                if (first_PSV(poslcp, lcp_value, i, value, cvalue - alph, first, lcpvalue, csa, lcp)){
                    return *poslcp + 1;
		}
            }
        }

        if (lcpvalue < value)//first not included in min of rule
	  return first + 1;
	
        if (pos == 0) return 0; //no rules to the left
    
        //move to the left
        cvalue = get_C(pos - 1);
        if (cvalue >= alph){//rule
            first -= rules->get_nt(cvalue - alph);
	    lcpvalue -= rules->get_sum(cvalue - alph);
	}
        else{
            --first;
	     lcpvalue -= cvalue;
	}
	
	//lcpvalue = lcp->get_LCP(first, csa);
	cvalue = get_C(pos - 1);
        for (k = pos - 1; k > 0; --k) {
            if (cvalue >= alph) {//rule
                min = rules->get_min(cvalue - alph) + lcpvalue;  
                if (min < value) { 
                    PSV_rule(poslcp, lcp_value, value, cvalue, first, lcpvalue, csa, lcp);
                    return *poslcp + 1;
                }

            } else {
                min = cvalue + lcpvalue;
                if (min < value) {
                    *lcp_value = min;
                    *poslcp = first + 1;
                    return *poslcp + 1;
                }
            }
	    if (lcpvalue < value)//first not included in min of rule
		return first + 1; 
            //move to the left
            cvalue = get_C(k - 1);
            if (cvalue >= alph) {//rule
                first -= rules->get_nt(cvalue - alph);
                lcpvalue -= rules->get_sum(cvalue - alph);
            } else {
                --first;
                lcpvalue -= cvalue;
            }
        }
        //k = 0
	if (cvalue >= alph) {//rule
                min = rules->get_min(cvalue - alph) + lcpvalue;
                if (min < value) {
                    PSV_rule(poslcp, lcp_value, value, cvalue, first, lcpvalue,  csa, lcp);
                    return *poslcp + 1;
                }

            } else {
                min = cvalue + lcpvalue;
                if (min < value) {
                    *lcp_value = min;
                    *poslcp = first+1;
                    return *poslcp + 1;
                }
            }
	
        return 0;
    }

    size_t NPR_RPBal::find_PSV(size_t i, RLCSA *csa, LCP_RLCSA *lcp) const {
        int poslcp, lcp_value, psv;
        psv = find_PSV(i, &poslcp, &lcp_value, csa, lcp);
        return psv;

    }

    size_t NPR_RPBal::bwd_PSV(size_t i, RLCSA *csa, LCP_RLCSA *lcp, size_t d) const {
        return 0;
    }

    size_t NPR_RPBal::fwd_NSV(size_t i, RLCSA *csa, LCP_RLCSA *lcp, size_t d) const {
        return 0;
    }
  
    int NPR_RPBal::RMQ_right(int *lcp_pos, int *lcp_value, int y, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const {
        //can asume min < value is in this tree
        int aux, left, right, cover, min, minpos, lcpvalue, lcppos, sum, rlcpvalue, rfirst, lmin,rmin, lpos, rpos;	

        if (rule < alph) {//terminal 
            *lcp_pos = first + 1;
            *lcp_value = first_lcp + rule;
            return *lcp_value;
        }
	
	
        cover = rules->get_nt(rule - alph);
	min = rules->get_min(rule - alph) + first_lcp;
	minpos = rules->get_lminpos(rule - alph) + first;
	if(y >= minpos){
		*lcp_pos = minpos;
		*lcp_value = min;
		return *lcp_value;
	}

	lcpvalue = first_lcp;
	lcppos = first;
        if (rule - alph >= trules) { //no children
            for (int i = first + 1; i <= y; ++i){
                if ((aux = lcp->get_LCP(i, csa)) < lcpvalue) {
                    lcpvalue = aux;
                    lcppos = i;
		    
                }
	     }
	     *lcp_pos = lcppos;
	     *lcp_value = lcpvalue;
	     return lcpvalue;
        }
	//printf("NSV_RULE value=%i, rule=%i, first=%i, cover=%i,\n",value,rule,first,cover);
        left = rules->get_left(rule - alph);
        right = rules->get_right(rule - alph);

        lcpvalue = first_lcp;
        if (left >= alph) { //rule
            min = rules->get_min(left - alph) + lcpvalue;
            cover = rules->get_nt(left - alph);
	    sum = rules->get_sum(left - alph);
        } else { //terminal
            min = lcpvalue;
            cover = 1;
	    sum = left;
        }
	rfirst = first + cover;
	rlcpvalue = lcpvalue + sum;
	
        if (y > rfirst){
		lmin = RMQ_right(&lpos, lcp_value, y, left, first, first_lcp, csa, lcp);			
		rmin = RMQ_right(&rpos, lcp_value, y, right, rfirst, rlcpvalue, csa, lcp);
		if(lmin <= rlcpvalue){
			*lcp_pos = lpos;
			return lmin;
		}else if(rlcpvalue <= rmin){
			*lcp_pos = rfirst;
			return rlcpvalue;
		}else{	
			*lcp_pos = rpos;	
			return rmin;
		}	
	}else{
		rmin = RMQ_right(lcp_pos, lcp_value, y, left, first, first_lcp, csa, lcp);
		if(rlcpvalue <= rmin){
			*lcp_pos = rfirst;
			return rlcpvalue;
		}else{
			return rmin;
		}
	}
    }	

	
	
     int NPR_RPBal::RMQ_left(int *lcp_pos, int *lcp_value, int x, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const {
        //can asume min < value is in this tree
        int aux,left, right, cover, min, minpos, lcpvalue, lcppos, sum, rlcpvalue, rfirst, lmin,rmin, lpos, rpos;	

        if (rule < alph) {//terminal 
            *lcp_pos = first + 1;
            *lcp_value = first_lcp + rule;
            return *lcp_value;
        }
	
	
        cover = rules->get_nt(rule - alph);
	min = rules->get_min(rule - alph) + first_lcp;
	minpos = rules->get_lminpos(rule - alph) + first;
	if(x <= minpos){
		*lcp_pos = minpos;
		*lcp_value = min;
		return *lcp_value;
	}

	lcpvalue = lcp->get_LCP(x, csa);
	lcppos = x;
        if (rule - alph >= trules) { //no children
            for (int i = x + 1; i <= first+cover; ++i){
                if ((aux = lcp->get_LCP(i, csa)) < lcpvalue) {
                    lcpvalue = aux;
                    lcppos = i;
		    
                }
	     }
	     *lcp_pos = lcppos;
	     *lcp_value = lcpvalue;
	     return lcpvalue;
        }
	//printf("NSV_RULE value=%i, rule=%i, first=%i, cover=%i,\n",value,rule,first,cover);
        left = rules->get_left(rule - alph);
        right = rules->get_right(rule - alph);

        lcpvalue = first_lcp;
        if (left >= alph) { //rule
            min = rules->get_min(left - alph) + lcpvalue;
            cover = rules->get_nt(left - alph);
	    sum = rules->get_sum(left - alph);
        } else { //terminal
            min = lcpvalue;
            cover = 1;
	    sum = left;
        }
	rfirst = first + cover;
	rlcpvalue = lcpvalue + sum;
	
        if (x <= rfirst){
		lmin = RMQ_left(&lpos, lcp_value, x, left, first, first_lcp, csa, lcp);	
		rmin = RMQ_left(&rpos, lcp_value, x, right, rfirst, rlcpvalue, csa, lcp);
		if(lmin <= rlcpvalue){
			*lcp_pos = lpos;
			return lmin;
		}else if(rlcpvalue <= rmin){
			*lcp_pos = rfirst;			
			return rlcpvalue;
		}else{	
			*lcp_pos = rpos;
			return rmin;
		}
	}else{
		rmin = RMQ_left(lcp_pos, lcp_value, x, right, rfirst, rlcpvalue, csa, lcp);
		return rmin;
	}
    }

    
    int NPR_RPBal::RMQ_rule(int *pos, int *lcp_value, size_t x, size_t y, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const {
        int cover, lcp_pos, aux, min_rule, minpos, lcpvalue, value;
	int left, right, min, sum, lmin, rmin, lpos, rpos, rlcpvalue, rfirst;
	//printf("rule_RMQ x=%i, y=%i\n",x,y );
	if(rule < alph){
	    *lcp_value = first_lcp;
	    *pos = first; 
	     return first;
	}

	min_rule = rules->get_min(rule - alph) + first_lcp;
	minpos = rules->get_lminpos(rule - alph) + first;
	if(x<= minpos && minpos<=y){
		*pos =  minpos;
	    	*lcp_value = min_rule;
	     	return *pos;
	}
	
	value = lcpvalue = lcp->get_LCP(x, csa);
	lcp_pos = x;


        cover = rules->get_nt(rule - alph);
        if (cover <= trim || rule - alph >= trules) { //no children, scan sequentially
            for (int i = x + 1; i <= y; ++i){
                if ((aux = lcp->get_LCP(i, csa)) < lcpvalue) {
                    lcpvalue = aux;
                    lcp_pos = i;
		    if(lcpvalue == min_rule) //reached min or rule_min
		    	break;
                }
	     }
	    *lcp_value = lcpvalue;
	    *pos = lcp_pos;
            return lcp_pos;
        }

	left = rules->get_left(rule - alph);
        right = rules->get_right(rule - alph);
	lcpvalue = first_lcp;
        if (left >= alph) { //rule
            min = rules->get_min(left - alph) + lcpvalue;
            cover = rules->get_nt(left - alph);
	    sum = rules->get_sum(left - alph);
        } else { //terminal
            min = lcpvalue;
            cover = 1;
	    sum = left;
        }
	rfirst = first + cover;
	rlcpvalue = lcpvalue + sum;

	lmin = RMQ_left(&lpos, lcp_value, x, left, first, first_lcp, csa, lcp);
        rmin = RMQ_right(&rpos, lcp_value, y, right, rfirst, rlcpvalue, csa, lcp);	
	
	if(lmin <= rmin)
		return lpos;
	else return rpos;

    }



  

    //Return the left-most position  j, x<= j <= y  , where is the minimun in the range [x,y]

    size_t NPR_RPBal::find_RMQ(size_t x, size_t y, RLCSA *csa, LCP_RLCSA *lcp) const {
        int pos_ini, pos_fin, first, first_ini, first_fin, aux, aux1, aux2, aux3, min, min_pos, cvalue, lcpvalue, first_lcp, last_lcp, pos;
        int lcp_pos, lcp_value, lcp_pos_fin, lcp_value_fin, fin_min;
	int ini_cover, first_min, first_min_lcp, ini_min, lcp_pos_ini, lcp_value_ini;
	min_pos = 0;
	//printf("find_RMQ x=%i, y=%i\n",x,y );
        if (x > y) {
            size_t z = y;
            y = x;
            x = z;
        }
        if (x == y) return x;

        min = len;
        pos_ini = pos_fin = first_ini = first_fin = 0;

        get_pos(x, &pos_ini, &first_ini, csa, lcp);
        get_pos(y, &pos_fin, &first_fin, csa, lcp);
	
	lcpvalue = first_lcp = lcp->get_LCP(first_ini, csa);
        if (pos_ini == pos_fin){//x,y are in same rule
	//printf("find_RMQ  lcp_pos_ini=%i, lcp_value=%i, x=%i, y=%i,get_C(pos_ini)=%i, first_ini=%i, lcpvalue=%i\n",lcp_pos_ini,lcp_value,x,y,get_C(pos_ini), first_ini, lcpvalue );
            return RMQ_rule(&lcp_pos, &lcp_value, x, y, get_C(pos_ini), first_ini, lcpvalue, csa, lcp);
	}

        first = first_ini;
        cvalue = get_C(pos_ini);
        if (cvalue >= alph) {//rule
            ini_min = rules->get_min(cvalue - alph) + lcpvalue;
	    ini_cover = rules->get_nt(cvalue - alph);	
            lcpvalue += rules->get_sum(cvalue - alph);
        } else {
            ini_min = cvalue + lcpvalue;
            ini_cover = 1;
            lcpvalue += cvalue;
        }
	first += ini_cover;

        for (int i = pos_ini + 1; i < pos_fin; ++i) {
            cvalue = get_C(i);
            if (cvalue >= alph) {//rule
                aux = rules->get_min(cvalue - alph) + lcpvalue;
                aux1 = rules->get_nt(cvalue - alph);
                aux2 = rules->get_sum(cvalue - alph);
		aux3 = rules->get_lminpos(cvalue - alph);
            } else {
                aux = cvalue + lcpvalue;
                aux1 = 1;
                aux2 = cvalue;
		aux3 = 1;
            }
	    if(lcpvalue < min){ //first value or rule
	    	min = lcpvalue;
		min_pos = first; 
	    }	
            if (aux < min) {
                min = aux;
		min_pos = aux3 + first;
                pos = i;
                first_min = first;
		first_min_lcp = lcpvalue;
            }
            first += aux1;
            lcpvalue += aux2;
        }
	last_lcp = lcpvalue;
        cvalue = get_C(pos_fin);
        if (cvalue >= alph)//rule
            fin_min = rules->get_min(cvalue - alph) + lcpvalue;
        else
            fin_min = cvalue + lcpvalue;

        //check first rule
	//ini_min could be to the left of x
        if (ini_min <= min) {               
	     int yy = RMQ_rule(&lcp_pos_ini, &lcp_value_ini, x, x+ini_cover, get_C(pos_ini), first_ini,first_lcp, csa, lcp);
	    //if(first_NSV(&lcp_pos_ini, &lcp_value_ini, x-1, min+1, get_C(pos_ini)-alph, first_ini, first_lcp, csa, lcp)){ 
		if (lcp_value_ini <= fin_min) {
		    return lcp_pos_ini;
		}
	//    }else{
	//	lcp_pos_ini = 0;
	//	lcp_value_ini = len;
	 //   }
	}
        //check last rule
	//fin_min could be to the right of y
        if (fin_min < min) {
	      RMQ_rule(&lcp_pos_fin, &lcp_value_fin, first_fin, y, get_C(pos_fin), first_fin, last_lcp, csa, lcp);
            //if(first_PSV(&lcp_pos_fin, &lcp_value_fin, y, min, get_C(pos_fin)-alph, first_fin, last_lcp, csa, lcp)){
		if (lcp_value_fin < lcp_value_ini) {
		    return lcp_pos_fin;
		}
	 //   }else{
	//	lcp_pos_fin = 0;
	//	lcp_value_fin = len;
	//    }
        }

        //check case lcp_value_ini>fin_min but lcp_value_ini<lcp_value_fin
        if (lcp_value_ini <= min && lcp_value_ini <= lcp_value_fin) {
            return lcp_pos_ini;
	    
        }else if (lcp_value_fin < min && lcp_value_fin < lcp_value) { 
            return lcp_pos_fin;
	    
        }else { 
		//NSV_rule(&lcp_pos, &lcp_value, min + 1, get_C(pos), first_min, first_min_lcp, csa, lcp); 
		return min_pos;
        }


    }

    size_t NPR_RPBal::getSize() const {
        size_t mem, Ssize;
        int Rsize, tsize;
        
	mem = 0;
        Ssize = floor((c - 1) / rate);
         //C[]
        mem += ((b * c + W - 1) / W) * sizeof (uint); //size_t mem = sizeof(size_t)*c;
        //length of rules array
        mem += rules->getSize();
        mem += sizeof (uint) * Ssize; //Sample (F)

        return mem;
    }

    void NPR_RPBal::save(ofstream & fp) const {
        size_t nb;

        saveValue(fp, npr_type);
        saveValue(fp, c);
        saveValue(fp, len);
        saveValue(fp, Cmin);
        saveValue(fp, n);
        saveValue(fp, alph);
        saveValue(fp, trim);
        saveValue(fp, rate);
        saveValue(fp, b);
        saveValue(fp, bmin);
	
	
	rules->save(fp);
        nb = (b * c + W - 1) / W; 
        saveValue(fp, C, (size_t) nb);
	nb = floor((c - 1) / rate);
        saveValue(fp, F, (size_t) nb); 
      //  nb = (bmin * c + W - 1) / W;
      //  saveValue(fp, min_pos, (size_t) nb); 
        

        printf("NPR saved\n");
    }
    NPR_RPBal * NPR_RPBal::load(ifstream & fp) {
        int nb, test;
	
        NPR_RPBal *npr = new NPR_RPBal();
        size_t type = loadValue<size_t > (fp);
        if (type != RPBal_NPR)
            abort();
        npr->c = loadValue<int > (fp); 
        npr->len = loadValue<int > (fp);
        npr->Cmin = loadValue<size_t> (fp);
        npr->n = loadValue<int > (fp);
        npr->alph = loadValue<int> (fp);
        npr->trim = loadValue<int > (fp);
        npr->rate = loadValue<int > (fp);
        npr->b = loadValue<size_t > (fp);
        npr->bmin = loadValue<size_t > (fp);
	//rules
        npr->rules = new Rules();
        npr->trules = npr->rules->load(fp);
        nb = (npr->b * npr->c + W - 1) / W; 
	test = nb;
        npr->C = loadValue<uint > (fp, (size_t) nb);
	nb = floor((npr->c - 1) / npr->rate);
        npr->F = loadValue<int>(fp, (size_t) nb);
        nb = (npr->bmin * npr->c + W - 1) / W;
        //npr->min_pos = loadValue<uint > (fp, (size_t) nb);

	cout << "alpha " << npr->alph << endl;
	cout << "trim size " << npr->rules->trim_size << endl;
	
	uint *C = new uint[npr->rules->size];
	for(int i=0; i<npr->rules->size; ++i)
		C[i] = npr->rules->get_nt(i);
	factorization *rep = new factorization(C, npr->rules->size);
	cout << "size DAC NT " << rep->getSize() << endl;
	
	
	uint *C4 = new uint[npr->rules->size];
	for(int i=0; i<npr->rules->size; ++i)
		C4[i] = npr->rules->get_min(i) + npr->rules->MinMax;
	rep = new factorization(C4, npr->rules->size);
	cout << "size DAC min " << rep->getSize() << endl;
	
	uint *C2 = new uint[npr->rules->trim_size];
	for(int i=0; i<npr->rules->trim_size; ++i)
		C2[i] = npr->rules->get_right(i) + npr->rules->Rmin;
	rep = new factorization(C2, npr->rules->trim_size);
	cout << "size DAC RIGHT " << rep->getSize() << endl;
	
	uint *C3 = new uint[npr->c];
	for(int i=0; i<npr->c; ++i)
		C3[i] = npr->get_C(i) + npr->Cmin;
	rep = new factorization(C3, npr->c);
	cout << "size C " << test*sizeof(uint) << endl;
	cout << "size DAC C " << rep->getSize() << endl;

	uint *C1 = new uint[npr->rules->size];
	for(int i=0; i<npr->rules->size; ++i)
		C1[i] = npr->rules->get_sum(i) + npr->rules->SumMax;
	rep = new factorization(C1, npr->rules->size);
	cout << "size DAC sum " << rep->getSize() << endl;

        return npr;
    }

    void NPR_RPBal::saveFullNPR(ofstream & fp) const {
        size_t nb;
        saveValue(fp, npr_type);
        saveValue(fp, c);
        saveValue(fp, len);
        saveValue(fp, Cmin);
        saveValue(fp, n);
        saveValue(fp, alph);
        saveValue(fp, b);
        nb = (b * c + W - 1) / W;
        saveValue(fp, C, (size_t) nb);


        //rules
        rules->save(fp);

        printf("NPR FULL saved\n");
    }

    NPR_RPBal * NPR_RPBal::loadNPR(ifstream & fp, int t) {
        size_t nb;
        NPR_RPBal *npr = new NPR_RPBal();

        size_t type = loadValue<size_t > (fp);
        if (type != RPBal_NPR)
            abort();
        npr->c = loadValue<int> (fp); 
        npr->len = loadValue<int> (fp);
        npr->Cmin = loadValue<size_t > (fp);
        npr->n = loadValue<int> (fp);
        npr->alph = loadValue<int > (fp);  
        npr->b = loadValue<size_t > (fp); 
        nb = (npr->b * npr->c + W - 1) / W;
        npr->C = loadValue<uint> (fp, (size_t) nb);    
        //rules
        npr->rules = new Rules();
        npr->trules = npr->rules->load(fp);

	int maxdepth = 0;
        for (int i = 0; i < npr->n - npr->alph; ++i) {
            if (maxdepth < npr->rules->get_nt(i))
                maxdepth = npr->rules->get_nt(i);
        }
	
        fprintf(stderr, "RePair succeeded\n\n");
        fprintf(stderr, "   Original ints: %i\n", npr->len);
        fprintf(stderr, "   Number of rules: %i\n", npr->n - npr->alph);
        fprintf(stderr, "   Max  rule cover: %i\n", maxdepth);
        fprintf(stderr, "   Final sequence length: %i\n", npr->c);
	
        npr->trim = npr->rate = t;
        npr->minsize = 2048;
        npr->factor = 0.85;

        npr->compress_CV2();
        deleted_rules = npr->trim_rule();
        npr->sampling(t);

        return npr;
    }

    NPR_RPBal::~NPR_RPBal() {

        Rec->~Records_Bal();
        hash->~Hash_Bal();
        heap->~HeapBal();
        delete [] L;
    }

}
