/*  
 * Copyright (C) 2012, Andres Abeliuk, all rights reserved.
 *
 * Andres Abeliuk <aabeliuk@dcc.uchile.cl>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef _NPR_RPBal_H
#define	_NPR_RPBal_H

#include "NPR.h"
#include "LCP_RLCSA.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <math.h>
#include "HashMap.h"

#include "Rules.h"

#include <Records_Bal.h>
#include <Hash_Bal.h>
#include <HeapBal.h>

#include <factorization.h>

/*
typedef struct _Rule
   { Tpair pair; // pair content
     int min; // min # of rule terminals
     // int max; // max # of rule terminals
     int nt; // terminals that rule represents
     int sum;// sum of terminals
   } Rule;
*/
namespace cds_static{
    /*****************************
    There are r=(log(n)/log(b))-1 levels
    ****************************/

    class NPR_RPBal: public NPR{
            private:
                    Tlist *L; // |L| = c;
                    Hash_Bal *hash; // hash table of pairs
                    HeapBal *heap; // special heap of pairs
                    Records_Bal *Rec; // records
                    float factor; // 1/extra space overhead; set closer to 1 for smaller and
                                          // slower execution
                    int minsize; // to avoid many reallocs at small sizes, should be ok as is
                    int u; // |text| and later current |C| with gaps
                    uint *C; // compressed text
                    uint *min_pos; // compressed text min positions
                    int c;  // real |C|
                    int alph; // max used terminal symbol
                    int n; // |R|
                    Rules *rules; //rules choseen
                    Rules *C_rules; //rules made artificially to compress C
                    int *F; // Sample of lcp position in compressed text
                    int *A;  // Sample of lcp Absolute values in compressed text
                    int rate; //sample rate
                    int len; // Text length
                    size_t b, bmin;
                    int trules; //# of rules not trim
                    size_t Cmin; //min value in C to represent negatives
                    NPR_RPBal();
                    void prepare ();
                    int repair ();
                    int expand (int i);
                    int trim_tree ()const;
                    int trim_rule ();
                    int get_C(size_t i) const;
                    void set_C(size_t i, size_t x) const;

                    void prnC ();
                    void prnRec ();
                    int first_NSV(int *pos, int *lcp_value, int i, int value, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const;
                    int first_PSV(int *pos, int *lcp_value, int i, int value, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp) const;
                    int NSV_rule(int *lcp_pos, int *lcp_value, int value, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const;
                    int PSV_rule(int *lcp_pos, int *lcp_value, int value, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const;
                    int LCP_seq(int first, int *lcp_pos, int *lcp_value, int value, int cover,RLCSA *csa, LCP_RLCSA *lcp)const;
                    int LCP_seq_left(int first, int *lcp_pos, int *lcp_value, int value, int cover,RLCSA *csa, LCP_RLCSA *lcp)const;
		    int RMQ_right(int *lcp_pos, int *lcp_value, int y, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const;
		    int RMQ_left(int *lcp_pos, int *lcp_value, int y, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const;	
                    int RMQ_rule(int *pos, int *lcp_value, size_t x, size_t y, int rule, int first, int first_lcp, RLCSA *csa, LCP_RLCSA *lcp)const;
                    void sampling(int rate);
                    void get_pos(int i, int *pos, int *first, RLCSA *csa, LCP_RLCSA *lcp)const;
                    int  BinarySearch(int i, int low, int high, int last)const;
                    void get_pos_seq(int i, int *pos, int *first, RLCSA *csa, LCP_RLCSA *lcp)const;
		    void test (int * a)const;
                    

            public:
                    int expandRule (int i, int *index, int *result)const;
                    int trim; // trimming size for rules (rules with less than 'trim' terminals are trimmed)
                    size_t npr_type;
                    /*NPR Data structure base on Canovas and Navarro "Practical Compressed Suffix Trees"
                     * @param lcp Any LCP array data structure
                     * @param block_Size Block size that will be used
                     * @param csa The compresses suffix array
                     */
                    NPR_RPBal(LCP_RLCSA *lcp, RLCSA *csa, int length, size_t block_Size, int _minsize, float _factor);
                    NPR_RPBal(LCP_RLCSA *lcp, RLCSA *csa, int length, int _minsize, float _factor);
                     

                    virtual ~NPR_RPBal();


                    // Return the position of NSV[i]
                    virtual size_t find_NSV(int i, int *poslcp, int *lcp_value, RLCSA *csa, LCP_RLCSA *lcp) const;
                    virtual size_t find_NSV(size_t i,  RLCSA *csa, LCP_RLCSA *lcp) const;

                    void compress_C();
                    void compress_CV2();

                    void get_min_pos(RLCSA *csa, LCP_RLCSA *lcp);

                    // Return the position of PSV[i]+1
                   virtual size_t find_PSV(int i, int *poslcp, int *lcp_value, RLCSA *csa, LCP_RLCSA *lcp) const;
                   virtual size_t find_PSV(size_t i,  RLCSA *csa, LCP_RLCSA *lcp) const;


                    virtual size_t bwd_PSV(size_t i, RLCSA *csa, LCP_RLCSA *lcp, size_t d) const;

                    virtual size_t fwd_NSV(size_t i, RLCSA *csa, LCP_RLCSA *lcp, size_t d) const;

                    //Return the left-most position  j, x<= j <= y  , where is the minimun in the range [x,y]
                    virtual size_t find_RMQ(size_t x, size_t y, RLCSA *csa, LCP_RLCSA *lcp) const;

                    virtual  size_t getSize() const;

                    virtual void save(ofstream & fp) const;
                    virtual void saveFullNPR(ofstream & fp) const;

                    static NPR_RPBal * load(ifstream & fp);
                    static NPR_RPBal * loadNPR(ifstream & fp, int trimm);

    };

};

#endif	/* _NPR_RPBal_H */


