 /*

Repair -- an implementation of Larsson and Moffat's compression and
decompression algorithms.
Copyright (C) 2010-current_year Gonzalo Navarro

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author's contact: Gonzalo Navarro, Dept. of Computer Science, University of
Chile. Blanco Encalada 2120, Santiago, Chile. gnavarro@dcc.uchile.cl

*/

#include "Records_Bal.h"
#include <stdio.h>

int NullFreq = 1 << (8*sizeof(int)-1);

Records_Bal::Records_Bal(float _factor, int _minsize)
{
    records = new Trecord[_minsize];
    maxsize = _minsize;
    size = 0;
    factor = _factor;
    minsize = _minsize;
    hash = NULL;
    heap = NULL;
    List = NULL;


}

int Records_Bal::insertRecord (Tpair pair )

   { int i,id;
     Trecord *rec;
     if (size == maxsize)
	{ maxsize /= factor;
	  //realloc
         Trecord *aux = new Trecord[maxsize];
         for(i=0; i<size; ++i)
             aux[i] = records[i];
         delete [] records;

         records = aux;  
	}
     int min, sum;
     sum = pair.left + pair.right;
     if(pair.left < sum)
         min = pair.left;
     else
         min = sum;
     
     id = size++;
     rec = &records[id];
     rec->pair = pair;
     rec->min = min;
     rec->nt = 2;
     hash->insertHash(id, rec);
     heap->insertHeap (id, rec);
     return id;
   }


int Records_Bal::insertRecordDiff (Tpair pair, int len)

   { int id, i;
     Trecord *rec;
     if (size == maxsize)
	{ maxsize /= factor;
          //realloc
         Trecord *aux = new Trecord[maxsize];
         for(i=0; i<size; ++i)
             aux[i] = records[i];
         delete [] records;

         records = aux;
	}
     int min,left,sum, leftmin, rightmin;
     left = pair.left - len;
     sum = left + pair.right - len;

     if(left < sum){
         min = left;
	 leftmin = rightmin = 1;
     }else if(left > sum){
         min = sum;
	 leftmin = rightmin = 2;
     }else{//case sum == left 
	 min = left;
	 leftmin = 1;
	 rightmin = 2;
     }
     
     
     id = size++;
     rec = &records[id];
     rec->pair = pair;
     rec->min = min;
     rec->minleftpos = leftmin;
     rec->minrightpos = rightmin;		
     rec->nt = 2;
     rec->sum = sum;
     hash->insertHash (id, records);
     heap->insertHeap (id, records);
     return id;
   }

int Records_Bal::insertRecordMin (Tpair pair, int min, int leftpos, int rightpos, int nt, int sum)

   { int id,i;
     Trecord *rec;
     if (size == maxsize)
	{ maxsize /= factor;
          //realloc
         Trecord *aux = new Trecord[maxsize];
         for(i=0; i<size; ++i)
             aux[i] = records[i];
         delete [] records;

         records = aux;
	}
     id = size++;
     rec = &records[id];
     rec->pair = pair;
     rec->min = min;
     rec->minleftpos = leftpos;
     rec->minrightpos = rightpos;		
     rec->nt = nt;
     rec->sum = sum;
     hash->insertHash (id, records);
     heap->insertHeap (id, records);
     return id;
   }

void Records_Bal::deleteRecord ()

   { int i;
     size--;
     if (size == 0)
        { maxsize = 0;
          delete [] records;
          records = NULL;
        }
     else if ((size < maxsize * factor * factor) &&
	      (size * factor >= minsize))
	{ maxsize *= factor;
	  //realloc
         Trecord *aux = new Trecord[maxsize];
         for(i=0; i<size; ++i)
             aux[i] = records[i];
         delete [] records;
         records = aux;
	}
   }


void Records_Bal::assocRecords (Hash_Bal *_hash, HeapBal *_heap, Tlist *list)

   { hash = _hash;
     heap = _heap;
     List = list;
   }

Records_Bal::~Records_Bal()
   { if (maxsize == 0) return;
     delete[] records;
     records = NULL;
     maxsize = 0;
     size = 0;
     hash = NULL;
     heap = NULL;
     List = NULL;
   }

void Records_Bal::removeRecord (int id) // delete record, freq <= 1
				       // due to freq 0 or purgue (freq 1)
				       // already deleted from heap

   { Tlist *L = List;
     hash->deleteHash (id, records); // mark del in hash
     if ((records[id].cpos != -1) && (L[records[id].cpos].prev == -id-1))
	L[records[id].cpos].prev = NullFreq; // null ptr from L
     if (id != size-1)
        { records[id] = records[size-1];
          hash->hashRepos (id, records);
          heap->heapRepos (id, records);
          L[records[id].cpos].prev = -id-1;
	}
     deleteRecord ();
   }


