/*

Repair -- an implementation of Larsson and Moffat's compression and
decompression algorithms.
Copyright (C) 2010-current_year Gonzalo Navarro

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author's contact: Gonzalo Navarro, Dept. of Computer Science, University of
Chile. Blanco Encalada 2120, Santiago, Chile. gnavarro@dcc.uchile.cl

*/

#ifndef _RECORDS_BAL_H
#define	_RECORDS_BAL_H

typedef struct _Tpair
  { int left,right;
  } Tpair;

typedef struct _Tlist
   { int prev,next;
   } Tlist; // list of prev next equal char


typedef struct _Trecord
   { Tpair pair; // pair content
     int freq; // frequency
     int cpos; // 1st position in C
     int hpos; // position in heap
     int kpos; // position in hash
     int min; // min # of rule terminals
     int minleftpos; // relative leftmost min position inside rule
     int minrightpos; // relative rightmost min position inside rule
     int nt; // terminals that rule represents
     int sum;// sum of terminals
   } Trecord;


   
typedef struct
   { Trecord *records;
     int maxsize;
     int size;
     float factor;
     int minsize;
     class Hash_Bal *Hash;  // Thash *
     class  HeapBal *Heap; // Theap *
     void *List; // Tlist *
   } Trarray;

   

class Records_Bal {
public:
     // contents can be accessed as Rec.records[0..Rec.size-1]
    Records_Bal(float factor, int minsize); // creates empty array
    int insertRecord (Tpair pair);
                            // inserts pair in Rec, returns id, links to/from
                            // Hash and Heap, not List. sets freq = 1
    int insertRecordDiff (Tpair pair, int len);
                            // inserts pair in Rec, returns id, links to/from
                            // Hash and Heap, not List. sets freq = 1
    int insertRecordMin(Tpair pair, int min, int leftpos, int rightpos, int nt, int sum);
    void deleteRecord (); // deletes last cell in Rec
    void assocRecords (Hash_Bal *hash, HeapBal *heap, Tlist *List); // associates structures
    void removeRecord (int id);// delete record, freq <= 1
                                           // due to freq 0 or purgue (freq 1)
                                           // already deleted from heap
    ~Records_Bal();// destroys Rec

    Trecord *records;
    int size;
    HeapBal *heap;
private:
     int maxsize;
     float factor;
     int minsize;
     Hash_Bal *hash; // Thash *
      // Theap *
     Tlist *List; // Tlist *


};

    #include "HeapBal.h"
    #include "Hash_Bal.h"

#endif	/* _RECORDS_H */

