/*  
 * Copyright (C) 2012, Andres Abeliuk, all rights reserved.
 *
 * Andres Abeliuk <aabeliuk@dcc.uchile.cl>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "Rules.h"


Rules::Rules()

   {
     size = 0;
     trim_size = 0;
     factor = 0;
     minsize = 0;
     maxsize = minsize;
     MinMax = SumMax = 0;
     Rmin = 0;
     b = 0;
     b2 = bnt = bmin = bsum = b;

     left = NULL;
     right = NULL;
     min = NULL;
     lminpos =rminpos = NULL;
     nt = NULL;
     sum = NULL;

   }


Rules::Rules(float _factor, int _minsize, int length)

   {
     size = 0;
     trim_size = 0;
     factor = _factor;
     minsize = _minsize;
     maxsize = minsize;
     MinMax = SumMax = length;
     Rmin = 0;
     b = bits(length-1)+1;
     b2 = bnt = bmin = bsum = b;
     long long nb = 1;
     nb = (nb*b*maxsize+W-1)/W;
     left = new uint[(size_t)nb];
     right = new uint[(size_t)nb];
     min = new uint[(size_t)nb];
     lminpos = new uint[(size_t)nb];
     rminpos = new uint[(size_t)nb];
     nt = new uint[(size_t)nb];
     sum = new uint[(size_t)nb];
     
     for(size_t i=0; i<(size_t)nb; i++){
         left[i] = 0;
         right[i] = 0;
         min[i] = 0;
	 lminpos[i] = 0;
	 rminpos[i] = 0;
         nt[i] = 0;
         sum[i] = 0;
     }

   }

void set_field64(uint *A, int len, size_t index, size_t x) {
        if(len==0) return;
        int i=index*len/W, j=index*len-i*W;
        uint mask = ((j+len) < W ? ~0u << (j+len) : 0)
            | ((W-j) < W ? ~0u >> (W-j) : 0);
        A[i] = (A[i] & mask) | x << j;
        if (j+len>W) {
            mask = ((~0u) << (len+j-W));
            A[i+1] = (A[i+1] & mask)| x >> (W-j);
        }
    }

    uint get_field64(uint *A, size_t len,  size_t index) {
        if(len==0) return 0;
        int i=index*len/W, j=index*len-W*i;
        uint result;
        if (j+len <= W)
            result = (A[i] << (W-j-len)) >> (W-len);
        else {
            result = A[i] >> j;
            result = result | (A[i+1] << (WW-j-len)) >> (W-len);
        }
        return result;
    }


void Rules::insertRule (int _left, int _right, int _min, int _lminpos, int _rminpos, int _nt, int _sum )
{
    uint *nleft, *nright, *nmin, *nlminpos, *nrminpos, *nnt, *nsum;
    size_t i, last_nb;

    if (size+1 == maxsize-1)
    {
       last_nb = (b*maxsize+W-1)/W;
       maxsize /= factor;
       long long nb = 1;
       nb = (nb*b*maxsize+W-1)/W;

       nleft = new uint[(size_t)nb];
       for (i=0;i<last_nb; i++)
           nleft[i] = left[i];
       for (i=last_nb; i<nb; i++)
           nleft[i] = 0;
       delete [] left;           // free old array memory.
       left = nleft;            // now a points to new array.

       nright = new uint[(size_t)nb];
       for (i=0;i<last_nb; i++)
           nright[i] = right[i];
       for (i=last_nb; i<nb; i++)
           nright[i] = 0;
       delete [] right;
       right = nright;

       nmin = new uint[(size_t)nb];
       for (i=0;i<last_nb; i++)
           nmin[i] = min[i];
       for (i=last_nb; i<nb; i++)
           nmin[i] = 0;
       delete [] min;
       min = nmin;

       nlminpos = new uint[(size_t)nb];
       nrminpos = new uint[(size_t)nb];
       for (i=0;i<last_nb; i++){
           nlminpos[i] = lminpos[i];
	   nrminpos[i] = rminpos[i];
       }
       for (i=last_nb; i<nb; i++){
           nlminpos[i] = 0;
	   nrminpos[i] = 0;
       }
       delete [] lminpos;
       delete [] rminpos;
       lminpos = nlminpos;
       rminpos = nrminpos;

       nnt = new uint[(size_t)nb];
       for (i=0;i<last_nb; i++)
           nnt[i] = nt[i];
       for (i=last_nb; i<nb; i++)
           nnt[i] = 0;
       delete [] nt;
       nt = nnt;

       nsum = new uint[(size_t)nb];
       for (i=0;i<last_nb; i++)
           nsum[i] = sum[i];
       for (i=last_nb; i<nb; i++)
           nsum[i] = 0;
       delete [] sum;
       sum = nsum;
       
    }
    
    set_field64(left, b, size, _left);
    set_field64(right, b, size, _right);
    set_field64(min, b, size, _min+MinMax);
    set_field64(lminpos, b, size, _lminpos);
    set_field64(rminpos, b, size, _rminpos);
    set_field64(nt, b, size, _nt);
    set_field64(sum, b, size, _sum+SumMax);
    size++;
    trim_size++;
    
}


void Rules::insertTrimRule (int _min, int _nt, int _lminpos, int _rminpos, int _sum )
{
    uint *nmin, *nlminpos, *nrminpos, *nnt, *nsum;
    size_t i, last_nb;

    if (size+1 == maxsize-1)
    {
       last_nb = (b*maxsize+W-1)/W;
       maxsize /= factor;
       long long nb = 1;
       nb = (nb*b*maxsize+W-1)/W;


       nmin = new uint[(size_t)nb];
       for (i=0;i<last_nb; i++)
           nmin[i] = min[i];
       for (i=last_nb; i<nb; i++)
           nmin[i] = 0;
       delete [] min;
       min = nmin;

       nlminpos = new uint[(size_t)nb];
       nrminpos = new uint[(size_t)nb];
       for (i=0;i<last_nb; i++){
           nlminpos[i] = lminpos[i];
	   nrminpos[i] = rminpos[i];
       }
       for (i=last_nb; i<nb; i++){
           nlminpos[i] = 0;
	   nrminpos[i] = 0;
       }
       delete [] lminpos;
       delete [] rminpos;
       lminpos = nlminpos;
       rminpos = nrminpos;

       nnt = new uint[(size_t)nb];
       for (i=0;i<last_nb; i++)
           nnt[i] = nt[i];
       for (i=last_nb; i<nb; i++)
           nnt[i] = 0;
       delete [] nt;
       nt = nnt;

       nsum = new uint[(size_t)nb];
       for (i=0;i<last_nb; i++)
           nsum[i] = sum[i];
       for (i=last_nb; i<nb; i++)
           nsum[i] = 0;
       delete [] sum;
       sum = nsum;

    }
    
    set_field64(min, b, size, _min+MinMax);
    set_field64(lminpos, b, size, _lminpos);
    set_field64(rminpos, b, size, _rminpos);
    set_field64(nt, b, size, _nt);
    set_field64(sum, b, size, _sum+SumMax);
    size++;

}



int Rules::get_left(int i){
    return get_field64(left, b2, i) - Rmin;
}
int Rules::get_right(int i){
    return get_field64(right, b2, i) - Rmin;
}
int Rules::get_min(int i){
    return get_field64(min, bmin, i) - MinMax;
}
int Rules::get_lminpos(int i){
    return get_field64(lminpos, bnt, i);
}
int Rules::get_rminpos(int i){
    return get_field64(rminpos, bnt, i);
}
int Rules::get_nt(int i){
    return get_field64(nt, bnt, i);
}
int Rules::get_sum(int i){
    return get_field64(sum, bsum, i) - SumMax;
}

void Rules::set_Rmin(int min){
    Rmin = min;

}

void Rules::set_left(int i, int value){
    return set_field64(left, b2, i, value);
}
void Rules::set_right(int i, int value){
    return set_field64(right, b2, i,value);
}

int* Rules::trim(HashMap *deleted, HashMap *trimed, Rules *Crules, int alph){
    size_t j, i, nsize, n, bb, b_nt, b_sum, b_min;
    int  nt_max, sum_max, sum_min, min_max, min_min, rule_max, aux, value;
    nt_max = sum_max = sum_min = min_max = min_min = rule_max = 0;
    //get max value to get # bits necesarry to represent
 
    for(i=0; i<size; ++i){
        
        aux = get_min(i);
        if(min_min>aux)
            min_min = aux;
        if(min_max<aux)
            min_max = aux;

        aux = get_sum(i);
        if(sum_min > aux)
           sum_min = aux;
        if(sum_max<aux)
            sum_max = aux;

        aux = get_nt(i);
        if(nt_max<aux)
            nt_max = aux;

        aux = get_left(i);
        if(rule_max<aux)
            rule_max = aux;
        aux = get_right(i);
        if(rule_max<aux)
            rule_max = aux;
    }
    for(i=0; i<Crules->size; ++i){
       
        aux = Crules->get_min(i);
        if(min_min>aux)
            min_min = aux;
        if(min_max<aux)
            min_max = aux;

        aux = Crules->get_sum(i);
        if(sum_min > aux)
           sum_min = aux;
        if(sum_max<aux)
            sum_max = aux;

        aux = Crules->get_nt(i);
        if(nt_max<aux)
            nt_max = aux;
    }
    min_min = -min_min;
    sum_min = -sum_min;
    
    nsize = size + Crules->size;
    int *location = new int[nsize];
    nsize -= deleted->used;

    b_nt = bits(nt_max+1);
    n = (b_nt*nsize+W-1)/W;
    uint *nnt = new uint[(size_t)n];
    for(size_t i=0; i<(size_t)n; i++)
         nnt[i] = 0;

    uint *nlminpos = new uint[(size_t)n];
    uint *nrminpos = new uint[(size_t)n];
    for(size_t i=0; i<(size_t)n; i++){
         nlminpos[i] = 0;
	 nrminpos[i] = 0;
    }

    b_min = bits(min_min+min_max+1);  
    n = (b_min*nsize+W-1)/W;
    uint *nmin = new uint[(size_t)n];
    for(size_t i=0; i<(size_t)n; i++)
         nmin[i] = 0;

    b_sum = bits(sum_min+sum_max+1);
    n = (b_sum*nsize+W-1)/W;
    uint *nsum = new uint[(size_t)n];
    for(size_t i=0; i<(size_t)n; i++)
         nsum[i] = 0;
    
    //trimed rules dont have son
   
    bb = bits(Rmin+rule_max+1);
    n = (bb*nsize+W-1)/W;
    nsize -=  trimed->used;
    nsize -=  Crules->size; //size of rules with son
   
    uint *nleft = new uint[(size_t)n];
    uint *nright = new uint[(size_t)n];
     for(size_t i=0; i<(size_t)n; i++){
         nleft[i] = 0;
         nright[i] = 0;
     }

    //reallocate
    j=0;
    for(i=0;i<size; ++i){
       if(deleted->get(i)!= -1){//erase
          location[i] = -1;

       }
       else if(trimed->get(i)!= -1){
          set_field64(nmin, b_min, nsize, get_min(i)+min_min);
	  set_field64(nlminpos, b_nt, nsize, get_lminpos(i));
	  set_field64(nrminpos, b_nt, nsize, get_rminpos(i));
          set_field64(nnt, b_nt, nsize, get_nt(i));
          set_field64(nsum, b_sum, nsize, get_sum(i)+sum_min);
          location[i] = nsize++;
       }
       else{
            set_field64(nleft, bb, j, get_left(i)+Rmin);
            set_field64(nright, bb,j, get_right(i)+Rmin);
            set_field64(nmin, b_min, j, get_min(i)+min_min);
	    set_field64(nlminpos, b_nt, j, get_lminpos(i));
	    set_field64(nrminpos, b_nt, j, get_rminpos(i));
            set_field64(nnt, b_nt, j, get_nt(i));
            set_field64(nsum, b_sum, j, get_sum(i)+sum_min);

            location[i] = j++; //name of rule has change
       }
    }

    for(i=0;i<Crules->size; ++i){
          set_field64(nmin, b_min, nsize, Crules->get_min(i)+min_min);
	  set_field64(nlminpos, b_nt, nsize, Crules->get_lminpos(i));
	  set_field64(nrminpos, b_nt, nsize, Crules->get_rminpos(i));
          set_field64(nnt, b_nt, nsize, Crules->get_nt(i));
          set_field64(nsum, b_sum, nsize, Crules->get_sum(i)+sum_min);
          location[i+size] = nsize++;
    }
       
    delete[] left;
    delete[] right;
    delete[] min;
    delete[] lminpos;
    delete[] rminpos;
    delete[] nt;
    delete[] sum;

    size = nsize;
    maxsize = size;
    if(j>0) trim_size = j-1;
    else trim_size=0;
    left = nleft;
    right = nright;
    min = nmin;
    lminpos = nlminpos;
    rminpos = nrminpos;
    nt = nnt;
    sum = nsum;
 
    b2 = bb;
    bnt = b_nt;
    bmin = b_min;
    bsum = b_sum;
    MinMax = min_min;
    SumMax = sum_min;

    
    //rename rules
    for(i=0;i<trim_size; ++i){
        value = get_left(i);
        if(value >= alph){
            value = location[value-alph];
            value += alph;
            set_field64(left, bb, i, value+Rmin);
        }
        value = get_right(i);
        if(value >= alph){
            value = location[value-alph];
            value += alph;
            set_field64(right, bb, i, value+Rmin);
        }
    }

   
    return location;
}

void Rules::save(ofstream & fp){
    int n;

    saveValue(fp, size);
    saveValue(fp, trim_size);
    saveValue(fp, b);
    saveValue(fp, b2);
    saveValue(fp, bnt);
    saveValue(fp, bsum);
    saveValue(fp, bmin);
    saveValue(fp, Rmin);
    saveValue(fp, MinMax);
    saveValue(fp, SumMax);     
    n = (b2*trim_size+W-1)/W;
    saveValue(fp, left, (size_t)n);
    saveValue(fp, right, (size_t)n);
    n = (bnt*size+W-1)/W;
    saveValue(fp, nt, (size_t)n);
    saveValue(fp, lminpos, (size_t)n);
    saveValue(fp, rminpos, (size_t)n);
    n = (bsum*size+W-1)/W;
    saveValue(fp, sum, (size_t)n);
    n = (bmin*size+W-1)/W;
    saveValue(fp, min, (size_t)n);

  
}

int Rules::load(ifstream & fp){
    int n;

    size = loadValue<int>(fp);
    trim_size = loadValue<int>(fp);
    b = loadValue<size_t>(fp);
    b2 = loadValue<size_t>(fp);
    bnt = loadValue<size_t>(fp);
    bsum = loadValue<size_t>(fp);
    bmin = loadValue<size_t>(fp);
    Rmin = loadValue<int>(fp);
    MinMax = loadValue<int>(fp);
    SumMax =loadValue<int>(fp); 
    n = (b2*trim_size+W-1)/W;
    left = loadValue<uint>(fp, (size_t)n);
    right = loadValue<uint>(fp, (size_t)n);
    n = (bnt*size+W-1)/W;
    nt = loadValue<uint>(fp, (size_t)n);
    lminpos = loadValue<uint>(fp, (size_t)n);
    rminpos = loadValue<uint>(fp, (size_t)n);
    n = (bsum*size+W-1)/W;
    sum = loadValue<uint>(fp, (size_t)n);
    n = (bmin*size+W-1)/W;
    min = loadValue<uint>(fp, (size_t)n);

    factor = 0;
    minsize = 0;
    maxsize = minsize;

	printf("Rules trim_size = %i\n",trim_size);
	printf("Rules size = %i\n",size);

    return trim_size;
}


size_t Rules::getSize(){
    size_t mem, n;
	factorization_var *nt_dac, *lminpos_dac; 
	factorization *rminpos_dac;

	//DAC
	uint *L = new uint[size];
	for(int i=0; i<size; ++i)
		L[i] = get_nt(i);
	nt_dac = new factorization_var(L, size);


	L = new uint[size];
	for(int i=0; i<size; ++i)
		L[i] = get_lminpos(i);
	lminpos_dac = new factorization_var(L, size);
	
	L = new uint[size];
	uint aux;
	for(int i=0; i<size; ++i){
		aux = get_rminpos(i) - get_lminpos(i);
		L[i] =  aux;
	}
	rminpos_dac = new factorization(L, size);

	//delete [] L;
	//minpos->~factorization_var();	
	
	mem = nt_dac->getSize();
	mem += lminpos_dac->getSize();
	mem += rminpos_dac->getSize();

	//nt_dac->~factorization_var();
	//minpos_dac->~factorization_var();

    n = (bnt*size+W-1)/W;
  //  printf("nt=%i\n",sizeof(size_t)*n*8);
	
//    mem = sizeof(uint)*n; //nt
//    mem += sizeof(uint)*n; //minpos

    n = (bmin*size+W-1)/W;
  //  printf("min=%i\n",sizeof(size_t)*n*8);
    mem += sizeof(uint)*n; //min
    n = (bsum*size+W-1)/W;
  //  printf("sum=%i\n",sizeof(size_t)*n*8);
    mem += sizeof(uint)*n; //sum
    n = (b*trim_size+W-1)/W;
  //  printf("RL=%i\n",sizeof(size_t)*n*16);
    mem += sizeof(uint)*n*2; //left, right

    return mem;
}


Rules::~Rules() {
     if (maxsize == 0) return;
     delete [] left;
     delete [] right;
     delete [] min;
     delete [] nt;
     delete [] sum;
     left = NULL;
     right = NULL;
     min = NULL;
     lminpos = NULL;
     rminpos = NULL;
     nt = NULL;
     sum = NULL;
     maxsize = 0;
     size = 0;
}

