/*  
 * Copyright (C) 2012, Andres Abeliuk, all rights reserved.
 *
 * Andres Abeliuk <aabeliuk@dcc.uchile.cl>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
#ifndef RULES_H
#define	RULES_H

#include <libcdsBasics.h>
#include <libcdsTrees.h>

#include "HashMap.h"

#include <factorization_var.h>
#include <factorization.h>

using namespace cds_utils;
using namespace cds_static;

class Rules {
public:
    Rules( );
    Rules(float _factor, int _minsize, int length);
    void insertRule (int _left, int _right, int _min, int _lminpos, int _rminpos, int _nt, int _sum );
    void insertTrimRule ( int _min, int _nt, int _lminpos, int _rminpos, int _sum );
    int get_left(int i);
    int get_right(int i);
    int get_min(int i);
    int get_lminpos(int i);
    int get_rminpos(int i);
    int get_nt(int i);
    int get_sum(int i);
    void set_left(int i, int value);
    void set_right(int i, int value);
    void set_Rmin(int min);
    int* trim(HashMap *deleted, HashMap *trimed, Rules *Crules, int alph);
    void save(ofstream & fp);
    int load(ifstream & fp);
int loadtemp(ifstream & fp);
    size_t getSize();
    virtual ~Rules();

    int size;
    int trim_size; //trim rules dont have left an right node
    int Rmin;
    
     int MinMax, SumMax; 
private:
   
   
    size_t b, b2, bnt, bsum, bmin;
    float factor;
    int minsize;
    int maxsize;

    uint *left; // left pair content
    uint *right; // right pair content
    uint *min; // min # of rule terminals
    uint *lminpos; // position of min relative to rule terminals
    uint *rminpos; // position of min relative to rule terminals
    uint *nt; // terminals that rule represents
    uint *sum;// sum of terminals

};

#endif	/* RULES_H */

