/* Copyright (C) 2010, Rodrigo Cánovas, all rights reserved.
 *
 *This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
#ifndef _SUFFIXTREE_RLCSA_H
#define	_SUFFIXTREE_RLCSA_H

#include "LCP_RLCSA.h"
#include "NPR_RPBal.h"
#include "NPR_CN.h"


namespace cds_static{

    static const size_t  CSTY = 1;

    class SuffixTREE_RLCSA {
            protected:
			size_t length;
            private:
                    SuffixTREE_RLCSA();

                   
                    NPR *npr;


            public:
                    RLCSA *csa;
                    LCP_RLCSA *lcp;
                    /*Create the compressed suffix tree data structure base on Canovas and Navarro "Practical Compressed Suffix Trees"
                     * @param text The text from where the suffix tree will be create
                     * @param n  The length of the text
                     * @param lcp_type The lcp data structure that will be use (NAIVE|SAD_GON_OS|FMN_RRR_OS|PT|PHI|DAC|DAC_VAR).
                     * @param npr_type The npr data structute that will be use (FMN_NPR|RPBal_NPR).
                     * @param b  The block size used in NPR
                     * */
                    SuffixTREE_RLCSA(char *text, size_t n, size_t lcp_type = SAD_GON_OS, size_t npr_type = RPBal_NPR, size_t b = 32);


                    /*Create the compressed suffix tree data structure base on Canovas and Navarro "Practical Compressed Suffix Trees"
                     *  @param _lcp The LCP array data structure
                     *  @param _npr The NPR data structure
                     *  @param csa  The Compressed suffix array data structure
                     *  */
                    SuffixTREE_RLCSA(LCP_RLCSA *_lcp, NPR_RPBal *_npr, RLCSA *csa);

                    SuffixTREE_RLCSA(LCP_RLCSA *_lcp, RLCSA *_csa, size_t npr_type, size_t b);
                    SuffixTREE_RLCSA(LCP_RLCSA *_lcp, RLCSA *_csa, size_t npr_type);//full NPR

                    /*Operations*/

                    //Return true is [vl,vr] is the root
                    virtual bool isRoot(size_t vl,size_t vr) const;

                    //Return, in [vl,vr], the Root
                     virtual void Root(size_t *vl, size_t *vr) const;

                     //Return vr-vl+1
                     virtual size_t Count(size_t vl, size_t vr) const;

                     //Return true if [wl,wr] is ancestor of [vl,vr]
                     virtual bool Ancestor(size_t wl, size_t wr, size_t vl, size_t vr) const;

                     //Return the String Depth of the node [vl,vr]
                     virtual size_t SDepth(size_t vl, size_t vr) const;

                     //If [vl,vr] is a leaf (vr == vl), then return the suxffix array value at position v_l, Otherwise NULL.
                     virtual size_t Locate(size_t vl, size_t vr) const;

                     //Return, in [parent_l, parent_r], the parent of the node [vl,vr]
                     virtual void Parent(size_t vl, size_t vr, size_t *parent_l, size_t *parent_r) const;

                     //Return, in [fchild_l, fchild_r], the first child of the node [vl,vr]
                     virtual void FChild(size_t vl, size_t vr, size_t *fchild_l, size_t *fchild_r) const;

                     //Return, in [nsibling_l, nsibling_r], the next sibling of the node [vl,vr]
                     virtual void NSibling(size_t vl, size_t vr, size_t *nsibling_l, size_t *nsibling_r) const;

                     //Return, in [slink_l, slink_r], the suffix link of the node [vl,vr]
                     virtual void SLink(size_t vl, size_t vr, size_t *slink_l, size_t *slink_r) const;

                     //Return, in [slinki_l, slinki_r], the suffix link applied i times over the node [vl,vr]
                     virtual void SLink_i(size_t vl, size_t vr, size_t i, size_t *slinki_l, size_t *slinki_r) const;

                     //Return, in [lca_l, lca_r], the lower common ancestor between nodes [vl,vr] and [wl,wr]
                     virtual void LCA(size_t vl, size_t vr, size_t wl, size_t wr, size_t *lca_l, size_t *lca_r) const;

                     //Return, in [child_l, child_r], the child of the node [vl,vr] which label path start with the char a
                     virtual void Child(size_t vl, size_t vr, char a, size_t *child_l, size_t *child_r) const;

                     //Return the i-th letter of the label-path of the node [vl,vr]
                     virtual char Letter(size_t vl, size_t vr, int i) const;

                     //Return the tree depth of the node [vl,vr]
                     virtual size_t TDepth(size_t vl, size_t vr) const;

                     //Return, in [laq_s_l, laq_s_r], the highest ancestor of [vl,vr] with string-depth >=d
                     virtual void LAQs(size_t vl, size_t vr, size_t d, size_t *laq_s_l, size_t *laq_s_r) const;

                     //Return, in [laq_t_l, laq_t_r], the highest ancestor of [vl,vr] with tree-depth >=d
                     virtual void LAQt(size_t vl, size_t vr, size_t d, size_t *laq_t_l, size_t *laq_t_r) const;

                     //Memory use in bytes
                     virtual size_t getSize() const;

		     //for testing
                      virtual size_t NSV(size_t i) const;
                      virtual size_t PSV(size_t i) const;
                      virtual size_t RMQ(size_t i, size_t j) const;
		      virtual size_t LCP(size_t i) const;
		      virtual int count_lcp();

                     virtual void save(ofstream & fp, char *text ) const;
                     virtual void saveNPR(ofstream & fp, char *text ) const;
                     virtual void saveSA_LCP(ofstream & fp, char *text ) const;

                    static SuffixTREE_RLCSA * load(ifstream & fp, char *textname );
                    static SuffixTREE_RLCSA * loadSA_LCP(ifstream & fp, char *textname );
                    static SuffixTREE_RLCSA * loadNPR(ifstream & fp, char *textname, int t );

                    virtual ~SuffixTREE_RLCSA();
    };

};


#endif	/* _SUFFIXTREE_RLCSA_H */


