/*

Repair -- an implementation of Larsson and Moffat's compression and
decompression algorithms.
Copyright (C) 2010-current_year Gonzalo Navarro

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author's contact: Gonzalo Navarro, Dept. of Computer Science, University of
Chile. Blanco Encalada 2120, Santiago, Chile. gnavarro@dcc.uchile.cl

*/


	// extendible array for pairs

#include <stdlib.h>
#include <stdio.h>
#include "array_RP.h"

array_RP::array_RP(Records *_Rec, float _factor, int _minsize)

   {
     factor = _factor;
     minsize = _minsize;
     Rec = _Rec;
     maxsize = minsize;
     pairs = new int[maxsize];
     size = 0;
     fst = 0;
     
   }

int array_RP::insertArray (int pair)
{
    int *npairs;
    int max,i,pos,id;

    if(maxsize == 0){
        maxsize = minsize;
        pairs = new int[maxsize];
    }
    if (size == maxsize)
    {//realloc
       max = maxsize;
       maxsize /= factor;
       npairs = new int[maxsize];
       for (i=0;i<size;i++)
       {
            id = pairs[fst]; 
            npairs[i] = id;
            Rec->records[id].hpos = i;
            fst = (fst+1) % max;
       }
     delete [] pairs;           // free old array memory.
     pairs = npairs;            // now a points to new array.
     fst = 0;

    }
    pos = (fst + size) % maxsize;
    pairs[pos] = pair;
    size++;
    return pos;
}

void array_RP::deleteArray ()

   { int *npairs;
     int i,id,max;
     fst = (fst+1) % maxsize;
     size--;
     if (size == 0)
	{ maxsize = 0;
	  delete [] pairs;
	  pairs = NULL;
          fst = 0;
	}
     else if ((size < maxsize * factor * factor) &&
	      (maxsize * factor >= minsize))
	{ max = maxsize;
          maxsize *= factor;
          npairs = new int[maxsize];
          for (i=0;i<size;i++)
	      { id = pairs[fst];
		npairs[i] = id;
		Rec->records[id].hpos = i;
		fst = (fst+1) % max;
	      }
	 delete [] pairs;           // free old array memory.
         pairs = npairs;            // now a points to new array.
         fst = 0;
	}
   }


array_RP::~array_RP()

   { if (maxsize == 0) return;
     delete [] pairs;
     pairs = NULL;
     Rec = NULL;
     maxsize = 0;
     size = 0;
   }

